
import scipy
import numpy as np

def sparse_var(E, axis=0):
    """calculate variance across the specified axis of a sparse matrix."""

    mean_gene = E.mean(axis=axis).A.squeeze()
    tmp = E.copy()
    tmp.data **= 2
    return tmp.mean(axis=axis).A.squeeze() - mean_gene ** 2

def sparse_rowwise_multiply(E, a):
    ''' multiply each row of sparse matrix by a scalar '''

    nrow = E.shape[0]
    w = scipy.sparse.lil_matrix((nrow, nrow))
    w.setdiag(a)
    return w * E

def mean_center(E, column_means=None):
    """mean-center columns of a sparse matrix"""

    if column_means is None:
        column_means = E.mean(axis=0)
    return E - column_means

def normalize_variance(E, column_stdevs=None):
    ''' variance-normalize columns of a sparse matrix '''

    if column_stdevs is None:
        column_stdevs = np.sqrt(sparse_var(E, axis=0))
    return sparse_rowwise_multiply(E.T, 1 / column_stdevs).T

def sparse_zscore(E, gene_mean=None, gene_stdev=None):
    """z-score normalize each column of a sparse matrix"""
    if gene_mean is None:
        gene_mean = E.mean(0)
    if gene_stdev is None:
        gene_stdev = np.sqrt(sparse_var(E))
    return sparse_rowwise_multiply((E - gene_mean).T, 1/gene_stdev).T




import ABCParse
import pickle
import pathlib
import pandas as pd
from typing import Dict


class InVitroFatePredictionSubsets(ABCParse.ABCParse):
    """
    Dictionary of train/test split subsets (indices and tables) for the
    LARRY in vitro fate prediction task.
    """

    def __init__(self, disable: bool = False, *args, **kwargs):
        
        file = pathlib.Path(__file__)
        fname = "FatePredictionSubsets.pkl"
        path = file.parent.joinpath(fname)
        
        self.__parse__(locals(), public = [None])
        
        if not self._disable:
            self._f = self._load()

    def _load(self) -> Dict:
        try:
            return pickle.load(self._path.open("rb"))
        except:
            return pd.read_pickle(self._path)

    @property
    def indices(self) -> Dict:
        return self._f["Indices"]

    @property
    def tables(self) -> Dict:
        return self._f["Tables"]



import pandas as pd


def row_norm_df(df: pd.DataFrame) -> pd.DataFrame:
    """
    Sum normalize a pandas DataFrame across rows.
    
    Parameters
    ----------
    df: pd.DataFrame
        Input (raw, un-normalized) df
    
    Returns
    -------
    df_norm: pd.DataFrame
        Row-sum normalized df
    """
    return df.div(df.sum(axis=1), axis=0)


def col_norm_df(df) -> pd.DataFrame:
    """
    Sum normalize a pandas DataFrame over columns.
    
    Parameters
    ----------
    df: pd.DataFrame
        Input (raw, un-normalized) df
    
    Returns
    -------
    df_norm: pd.DataFrame
        Column-sum normalized df
    """
    return df.div(df.sum(axis=0), axis=1)




from ._cell_cycle_genes import cell_cycle_genes
from ._running_quantile import RunningQuantile
from ._vscores import vscores
from ._highly_variable_genes import highly_variable_genes
from ._remove_cell_cycle_correlated_genes import remove_cell_cycle_correlated_genes




from ._multi_fate_in_vitro_color_map import MultiFateInVitroColorMap, multi_fate_in_vitro_cmap
from ._in_vitro_color_map import InVitroColorMap, in_vitro_cmap
from ._confusion_matrix import ConfusionMatrix, confusion_matrix




import ABCParse

class LARRY_t0_test_indices(ABCParse.ABCParse):
    def __init__(
        self,
        adata,
        time_key="Time point",
        t0=2,
        lineage_key="clone_idx",
        test_key="test",
    ):

        # could generalize to a more iterative "subsetting".... rather than specified (time, lineage, test)

        self.__parse__(locals())

        self.df = self.adata.obs.copy()

    @property
    def future_time(self):
        _t = self.df[self.time_key].unique()
        return _t[_t != 2].tolist()

    @property
    def test_lineages(self):
        return (
            self.df.loc[self.df[self.test_key]]
            .loc[self.df[self.time_key].isin(self.future_time)]
            .loc[self.df[self.lineage_key].notna()]["clone_idx"]
        )

    @property
    def t0_obs(self):
        return self.df.loc[self.df[self.time_key] == self.t0]

    def __call__(self, sample=1):
        return (
            self.df.loc[self.df[self.lineage_key].isin(self.test_lineages)]
            .sample(sample, replace=False)
            .index
        )



# import packages: ------------------------------------------------------------
# import vinplots
import numpy as np
import matplotlib.pyplot as plt


# import local dependencies: --------------------------------------------------
from ._count_clonal_lineages import count_clonal_lineages


# supporting functions: DaughterCellPlots -------------------------------------
class DaughterCellPlots:
    def __init__(self):
        ...
        
    def mk_plot(self):

        fig = vinplots.Plot()
        fig.construct(
            nplots=4,
            ncols=2,
            figsize_height=0.8,
            figsize_width=1.2,
            hspace=0.4,
            wspace=0.2,
            height_ratios=[0.3, 0.7],
        )
        axes = fig.linearize()
        fig.modify_spines(ax=axes[0], spines_to_delete=["top", "right"])

        axes[0].set_title("Daughter cells per d2 clone", fontsize=10)
        axes[0].set_ylabel("Count", fontsize=8)

        for ax in [axes[2], axes[3]]:
            fig.modify_spines(
                ax=ax, spines_to_delete=["top", "right", "bottom", "left"]
            )
            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_xlabel("UMAP-1", fontsize=8)
            ax.set_ylabel("UMAP-2", fontsize=8)

        self.fig = fig
        self.axes = axes
        self.axes[1].remove()

    def plot_hist(self, n_daughters, bins=50):
        self.hist_bins = self.axes[0].hist(
            n_daughters,
            bins=bins,
            range=(1, n_daughters.max()),
            color="dodgerblue",
            edgecolor="k",
            log=True,
        )

    def plot_umap(self, adata, d2_mask, count, axn=2, day=4):

        X_umap = adata.obsm["X_umap"]
        meta_df = adata.obs.copy()

        self.axes[axn].scatter(X_umap[:, 0], X_umap[:, 1], s=5, color="lightgrey")
        d2_idx = meta_df.loc[meta_df["Time point"] == 2][d2_mask].index.astype(int)

        xu = X_umap[d2_idx]
        c = count[day] / count[2]
        c_idx = np.argsort(c)
        img = self.axes[axn].scatter(
            xu[c_idx, 0],
            xu[c_idx, 1],
            s=5,
            c=c[c_idx],
            vmax=8,
            cmap=plt.cm.plasma,
        )
        self.axes[axn].set_title("Day {} daughters counts".format(day), fontsize=10)
        plt.colorbar(mappable=img, ax=self.axes[axn], shrink=0.5)

# supporting functions: count daughter cells ----------------------------------
def _is_d2(obs_df, time_key):
    return obs_df[time_key] == 2


def _is_d2_clone(adata, time_key="Time point"):

    X_clone = adata.obsm["X_clone"]
    obs_df = adata.obs.copy()

    is_d2 = _is_d2(obs_df, time_key)
    is_d2_clone = (X_clone[is_d2, :].sum(axis=0) > 0).A.flatten()
    adata.uns["is_d2_clone"] = is_d2_clone
    adata.uns["n_d2_clones"] = is_d2_clone.sum()
    
def _count_daughter_cells(adata):
    adata.uns["n_daughters"] = adata.obsm["X_clone"][:, adata.uns["is_d2_clone"]].sum(0).A.flatten()


def _is_np_array(x):
    return x.__class__ is np.ndarray


def _count_d2_daughter_cells(adata):

    """
    For each d2 clone, count num. daughter cells.
    """

    meta_df = adata.obs.copy()

    d2_meta = meta_df.loc[meta_df["Time point"] == 2]
    d2_mask = d2_meta["clone_idx"].notna().values
    
    lineage_count_df = adata.uns['lineage_count_df']
    
    t = np.sort(meta_df["Time point"].unique())

    count = {}
    for d in t:
        count[d] = lineage_count_df.iloc[d2_meta[d2_mask]["clone_idx"]][d].values

    adata.uns["n_d2_daughters"], adata.uns["d2_lin_mask"] = count, d2_mask


# primary function: -----------------------------------------------------------
def count_d2_daughter_cells(adata, bins=50, plot_hist=True, plot_UMAPs=True):

    """plot histogram and umap of n_daughters"""
    
    _is_d2_clone(adata)
    _count_daughter_cells(adata)
    _count_d2_daughter_cells(adata)

    dcplot = DaughterCellPlots()
    dcplot.mk_plot()
    if plot_hist:
        dcplot.plot_hist(adata.uns["n_daughters"], bins=bins)
    if plot_UMAPs:
        dcplot.plot_umap(
            adata, adata.uns["d2_lin_mask"], adata.uns["n_d2_daughters"], axn=2, day=4
        )
        dcplot.plot_umap(
            adata, adata.uns["d2_lin_mask"], adata.uns["n_d2_daughters"], axn=3, day=6
        )



import numpy as np


from ._vscores import vscores

def highly_variable_genes(
    adata,
    base_ix=[],
    min_vscore_pctl=85,
    min_counts=3,
    min_cells=3,
    show_vscore_plot=False,
    sample_name="",
    return_idx = False,
):
    """
    Filter genes by expression level and variability
    Return list of filtered gene indices
    """
    
    E = adata.X

    if len(base_ix) == 0:
        base_ix = np.arange(E.shape[0])

    Vscores, CV_eff, CV_input, gene_ix, mu_gene, FF_gene, a, b = vscores(
        E[base_ix, :]
    )
    ix2 = Vscores > 0
    Vscores = Vscores[ix2]
    gene_ix = gene_ix[ix2]
    mu_gene = mu_gene[ix2]
    FF_gene = FF_gene[ix2]
    min_vscore = np.percentile(Vscores, min_vscore_pctl)
    ix = ((E[:, gene_ix] >= min_counts).sum(0).A.squeeze() >= min_cells) & (
        Vscores >= min_vscore
    )
    
    hv_idx = gene_ix[ix]

    
    adata.var["hv_genes"] = adata.var.index.astype(int).isin(hv_idx)
    
    if return_idx:
        return hv_idx



# -- import local dependencies: -----------------------------------------------
from ._interpolation_data import InterpolationData


# -- import packages: ---------------------------------------------------------
import ABCParse
import autodevice
import torch


# -- set typing: --------------------------------------------------------------
from typing import Tuple


# -- Operational class: -------------------------------------------------------
class InterpolationTask(ABCParse.ABCParse):
    def __init__(
        self,
        adata,
        time_key="Time point",
        use_key="X_pca",
        t0=2,
        n_samples=10_000,
        lineage_key="clone_idx",
        device=autodevice.AutoDevice(),
        backend = "auto",
        silent = False,
        PCA = None,
        *args,
        **kwargs,
    ):
        self.__parse__(locals())

        self.data = InterpolationData(**self._DATA_KWARGS)
        
        from scdiffeq.core.lightning_models.base import SinkhornDivergence
        self._sinkhorn_fn = SinkhornDivergence
        
        self.SinkhornDivergence = self._sinkhorn_fn(**self._SINKHORN_KWARGS)

    @property
    def _DATA_KWARGS(self):
        return ABCParse.function_kwargs(
            func=InterpolationData, kwargs=self._PARAMS
        )
    @property
    def _SINKHORN_KWARGS(self):
        return ABCParse.function_kwargs(
            func=self._sinkhorn_fn, kwargs=self._PARAMS
        )

    def forward_without_grad(self, DiffEq):
        """Forward integrate over the model without gradients."""
        with torch.no_grad():
            X_hat = DiffEq.forward(self.data.X0, self.data.t)
            return self._parse_forward_out(X_hat)

    def forward_with_grad(self, DiffEq):
        """Forward integrate over the model retaining gradients."""
        torch.set_grad_enabled(True)
        X_hat = DiffEq.forward(self.data.X0, self.data.t)
        return self._parse_forward_out(X_hat)
    
    @property
    def potential(self):
        return "Potential" in str(self._DiffEq)
    
    def _parse_forward_out(self, X_hat):
        """to account for KLDiv"""
        if isinstance(X_hat, Tuple):
            return X_hat[0]
        return X_hat
    
    def _dimension_reduce_pca(self, X_hat):
        return torch.stack(
            [torch.Tensor(self._PCA.transform(x)) for x in X_hat.detach().cpu().numpy()]
        ).to(self.device)     


    def __call__(self, trainer, DiffEq, *args, **kwargs):
        
        self.__update__(locals())
        
        if self.potential:
            X_hat = self.forward_with_grad(DiffEq)
        else:
            X_hat = self.forward_without_grad(DiffEq)
            
        if not self._PCA is None:
            X_hat = self._dimension_reduce_pca(X_hat)
        
        d4_loss = self.SinkhornDivergence(X_hat[1], self.data.X_test_d4).item()
        d6_loss = self.SinkhornDivergence(X_hat[2], self.data.X_train_d6).item()

        if not self._silent:
            print(
                "- Epoch: {:<5}| Day 4 loss: {:.2f} | Day 6 loss: {:.2f}".format(
                    DiffEq.current_epoch, d4_loss, d6_loss,
                ),
            )

        return d4_loss, d6_loss




__module_name__ = "_calculate_dominate_fate.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# import packages: ------------------------------------------------------------
from tqdm.notebook import tqdm
import pandas as pd
import numpy as np
import anndata


# import local dependencies: --------------------------------------------------
from ._get_lineage_obs import get_lineage_obs


# supporting functions: -------------------------------------------------------
def _compose_major_fate_df(d2_clone_indices, idx_max_list):
    return pd.DataFrame(
        {"clone_idx": d2_clone_indices, "d4_d6_major_fate_annot": idx_max_list}
    )


# -----------------------------------------------------------------------------
def calculate_dominate_fate(
    adata: anndata.AnnData, t_query: list = [4, 6], return_df: bool = False,
) -> pd.DataFrame:

    """
    Parameters:
    -----------
    adata
        type: anndata.AnnData

    t_query
        type: list
        default: [4, 6]

    Returns:
    --------
    major_fate_df
        type: pandas.DataFrame

    Notes:
    ------

    """

    d2_clone_indices = np.where(adata.uns["is_d2_clone"])[0]

    idx_max_list = []
    for lin_idx in tqdm(d2_clone_indices):
        df = get_lineage_obs(adata, lin_idx)
        idx_max_list.append(df.loc[df.index.isin(t_query)].sum(0).idxmax())
        
    major_fate_df = _compose_major_fate_df(d2_clone_indices, idx_max_list)
        
    adata.uns['major_fate_df'] = major_fate_df
        
    if return_df:
        return major_fate_df



import ABCParse
import pathlib
import numpy as np
import pandas as pd
import anndata

from typing import Optional


class WeinrebTestSet(ABCParse.ABCParse):
    """335 cell test set used in Weinreb, 2020 and Yeo, 2021"""

    def __init__(self, *args, **kwargs):
        
        """
        Parameters
        ----------
        base_path: str
        
        """

        self.__parse__(locals())
        
    @property
    def _BASE_PATH(self):
        
        self._sub_path = "assets/files/Figure5_Weinreb2020"
        return list(pathlib.Path(__file__).parents)[3].joinpath(self._sub_path)

    @property
    def _TIME_POINTS(self):
        if not hasattr(self, "_time_points"):
            PATH = self._BASE_PATH.joinpath("timepoints.npy")
            self._time_points = np.load(PATH)
        return self._time_points

    @property
    def _NM_MASK(self):
        if not hasattr(self, "_nm_mask"):
            PATH = self._BASE_PATH.joinpath("neutrophil_monocyte_trajectory_mask.npy")
            self._nm_mask = np.load(PATH)
        return self._nm_mask

    @property
    def _CLONAL_FATE_MATRIX(self):
        if not hasattr(self, "_clonal_fate_matrix"):
            PATH = self._BASE_PATH.parent.parent.parent.joinpath(
                "fate_bias_files/clonal_fate_matrix.npy"
            )
            self._clonal_fate_matrix = np.load(PATH)
        return self._clonal_fate_matrix

    @property
    def _PBA_predictions(self):
        if not hasattr(self, "_pba_preds"):
            PATH = self._BASE_PATH.joinpath("PBA_predictions.npy")
            self._pba_preds = np.load(PATH)
        return self._pba_preds

    @property
    def _FateID_predictions(self):
        if not hasattr(self, "_fateid_preds"):
            PATH = self._BASE_PATH.joinpath("FateID_predictions.npy")
            self._fateid_preds = np.load(PATH)
        return self._fateid_preds

    @property
    def _WOT_predictions(self):
        if not hasattr(self, "_wot_preds"):
            PATH = self._BASE_PATH.joinpath("WOT_predictions.npy")
            self._wot_preds = np.load(PATH)
        return self._wot_preds

    @property
    def _EARLY_CELLS(self):
        if not hasattr(self, "_early_cells"):
            PATH = self._BASE_PATH.joinpath("early_cells.npy")
            self._early_cells = np.load(PATH)
        return self._early_cells

    @property
    def _HELDOUT_MASK(self):
        if not hasattr(self, "_heldout_mask"):
            PATH = self._BASE_PATH.joinpath("heldout_mask.npy")
            self._heldout_mask = np.load(PATH)
        return self._heldout_mask

    @property
    def _SMOOTHED_GROUNDTRUTH(self):
        if not hasattr(self, "_smoothed_groundtruth"):
            PATH = self._BASE_PATH.joinpath("smoothed_groundtruth_from_heldout.npy")
            self._smoothed_groundtruth = np.load(PATH)
        return self._smoothed_groundtruth

    @property
    def _SMOOTHED_GROUNDTRUTH_SUBSET(self):
        if not hasattr(self, "_smoothed_groundtruth_subset"):
            self._smoothed_groundtruth_subset = self._SMOOTHED_GROUNDTRUTH[
                np.all(
                    [
                        self._EARLY_CELLS[self._NM_MASK[self._TIME_POINTS == 2]],
                        ~self._HELDOUT_MASK,
                    ],
                    axis=0,
                )
            ]
        return self._smoothed_groundtruth_subset

    @property
    def _NM_COUNTS(self):
        return self._CLONAL_FATE_MATRIX[:, 5:7].sum(1)

    @property
    def _HAS_FATE_MASK(self):
        return np.all(
            [self._NM_COUNTS > 0, self._NM_MASK[self._TIME_POINTS == 2]], axis=0
        )

    @property
    def _NM_PERCENT(self):
        return self._CLONAL_FATE_MATRIX[
            self._HAS_FATE_MASK, 5
        ] / self._CLONAL_FATE_MATRIX[self._HAS_FATE_MASK, 5:7].sum(1)

    @property
    def _HELDOUT_MASK_EARLY_CELLS(self):
        return (~self._HELDOUT_MASK)[
            self._EARLY_CELLS[self._NM_MASK[self._TIME_POINTS == 2]]
        ]

    def _subset_preds(self, method_prediction):
        return method_prediction[
            self._EARLY_CELLS[self._NM_MASK[self._TIME_POINTS == 2]]
        ][self._HELDOUT_MASK_EARLY_CELLS]

    @property
    def _PREDICTIONS(self):
        if not hasattr(self, "_predictions"):
            predictions = {
                "smoothed_ground_truth": self._SMOOTHED_GROUNDTRUTH_SUBSET,
            }
            for attr_name in self.__dir__():
                if "_predictions" in attr_name:
                    preds = getattr(self, attr_name)
                    predictions[attr_name.strip("_").split("_")[0]] = self._subset_preds(preds)
            self._predictions = predictions
        return self._predictions

    @property
    def _CLONAL_DATA_NO_HELDOUT(self):
        return self._NM_PERCENT[
            np.all(
                [
                    self._EARLY_CELLS[self._NM_MASK[self._TIME_POINTS == 2]],
                    ~self._HELDOUT_MASK,
                ],
                axis=0,
            )[self._HAS_FATE_MASK[self._NM_MASK[self._TIME_POINTS == 2]]]
        ]

    @property
    def _HAS_FATE_MASK_EARLY_HELDOUT(self):
        return self._HAS_FATE_MASK[
            np.all([self._NM_MASK[self._TIME_POINTS == 2], self._EARLY_CELLS], axis=0)
        ][self._HELDOUT_MASK_EARLY_CELLS]

    @property
    def _PREDICTIONS_FILTERED(self):
        if not hasattr(self, "_predictions_filtered"):
            self._predictions_filtered = {
                method: pred[self._HAS_FATE_MASK_EARLY_HELDOUT]
                for method, pred in self._PREDICTIONS.items()
            }
        return self._predictions_filtered

    @property
    def _D2_IDX(self):
        if hasattr(self, "_adata"):
            return self._adata.obs.loc[self._adata.obs["Time point"] == 2].index

    @property
    def _TEST_SET_IDX(self):
        if not self._D2_IDX is None:
            return self._D2_IDX[
                np.all(
                    [
                        self._NM_MASK[self._TIME_POINTS == 2],
                        self._EARLY_CELLS,
                    ],
                    axis=0,
                )
            ][self._HELDOUT_MASK_EARLY_CELLS][self._HAS_FATE_MASK_EARLY_HELDOUT]

    @property
    def df(self):
        if not hasattr(self, "_df"):
            COLS = [
                'clonal',
                'smoothed_ground_truth',
                'PBA',
                'FateID',
                'WOT',
            ]
            pred_dict = self._PREDICTIONS_FILTERED
            pred_dict["clonal"] = self._CLONAL_DATA_NO_HELDOUT
            self._df = pd.DataFrame(pred_dict, index=self._TEST_SET_IDX)[COLS]
            
        return self._df

    def __call__(self, adata: Optional[anndata.AnnData] = None, *args, **kwargs):
        
        """
        Parameters
        ----------
        adata: Optional[anndata.AnnData], default = None
        
        Returns
        -------
        test_set_df: pd.DataFrame
        """

        self.__update__(locals())

        return self.df

    
def weinreb_test_set(adata: Optional[anndata.AnnData] = None) -> pd.DataFrame:
    """
    Parameters
    ----------
    adata: Optional[anndata.AnnData], default = None

    Returns
    -------
    test_set_df: pd.DataFrame
    """
    
    test_set = WeinrebTestSet()

    return test_set(adata = adata)



# __init__.py for dataset
from ._figshare_downloader import FigshareDownloader

from ._figshare_downloads import (
    LARRY_timepoint_interpoltion as timepoint_interpolation,
    LARRY_fate_prediction as fate_prediction,
    LARRY_spliced_unspliced_reads as spliced_unspliced_reads,
)



# from ._dimension_reduction import DimensionReduction
from ._directory_manager import mkdir
from ._url_path_interfaces import (
    inVitroURLPaths,
    inVivoURLPaths,
    CytokinePerturbationURLPaths,
)
from ._load_expr_matrix import load_expr_matrix

from ._data import (
    inVitroData,
    inVivoData,
    CytokinePerturbationData,
)


from .klein_lab_pp_recipe import (
    RunningQuantile,
    cell_cycle_genes,
    vscores,
    highly_variable_genes,
    remove_cell_cycle_correlated_genes,
)

from ._split_data import (
    split_for_timepoint_recovery_task,
    split_for_fate_prediction_task,
    split_for_transfer_learning_task,
    SplitDataForTask,
)

from . import _dataset_utils as utils #  import Messages

from ._anndata_configuration import AnnDataConfiguration
from ._anndata_path_manager import AnnDataPathManager




import ABCParse

class DimensionReduction(ABCParse.ABCParse):
    def __init__(self, n_pcs=50, n_components=2, metric="euclidean", n_neighbors=30):

        self.__configure__(locals())
        
        import sklearn
        import umap

    def __configure__(self, kwargs, ignore=["self"]):

        self.__parse__(kwargs, ignore=ignore)

        self._scaler = sklearn.preprocessing.StandardScaler()
        self._pca_model = sklearn.decomposition.PCA(n_components=self.n_pcs)
        self._umap_model = umap.UMAP(
            n_components=self.n_components,
            metric=self.metric,
            n_neighbors=self.n_neighbors,
        )

    @property
    def Scaler(self):
        return self._scaler

    @property
    def PCA(self):
        return self._pca_model

    @property
    def UMAP(self):
        return self._umap_model



# -- import packages: ----------------------------------------------------
import pandas as pd
import numpy as np
import anndata
import time


# -- import local dependencies: ------------------------------------------
from ._dimension_reduction import DimensionReduction


# -- operator class: -----------------------------------------------------
class SplitDataForTask:
    def __init__(
        self,
        adata,
        split_key: str,
        train_vals,
        test_vals,
        n_pcs=50,
        n_components=2,
        metric="euclidean",
        n_neighbors=30,
    ):

        self.t0 = time.time()

        self.split_key = split_key
        self.train_vals = train_vals
        self.test_vals = test_vals
        self.adata = adata.copy()
        self.obs_df = self.adata.obs.copy()
        self.var_df = self.adata.var.copy()
        self.dim_reducer = DimensionReduction(
            n_pcs=n_pcs,
            n_components=n_components,
            metric=metric,
            n_neighbors=n_neighbors,
        )

    @property
    def train_idx(self):
        return self.obs_df.loc[self.obs_df[self.split_key].isin(self.train_vals)].index

    @property
    def test_idx(self):
        return self.obs_df.loc[self.obs_df[self.split_key].isin(self.test_vals)].index

    @property
    def X_train(self):
        if not hasattr(self, "_X_train"):
            self._X_train = self.adata[self.train_idx, self.var_df["use_genes"]].X.A
        return self._X_train
    
    def t_elapsed_message(self, message):
        msg = "- time elapsed: {:.2f}".format(time.time() - self.t0)
        print("{:<23} | {}".format(msg, message))

    @property
    def X_train_scaled(self):
        if not hasattr(self, "_X_train_scaled"):
            message = "Fitting scaling model on raw training data."
            self.t_elapsed_message(message)
            self._X_train_scaled = self.dim_reducer.Scaler.fit_transform(self.X_train)
        return self._X_train_scaled

    @property
    def X_train_pca(self):
        if not hasattr(self, "_X_train_pca"):
            message = "Fitting PCA model on scaled training data."
            self.t_elapsed_message(message)
            self._X_train_pca = self.dim_reducer.PCA.fit_transform(self.X_train_scaled)
        return self._X_train_pca

    @property
    def X_train_umap(self):
        if not hasattr(self, "_X_train_umap"):
            message = "Fitting UMAP model on training PCA projection."
            self.t_elapsed_message(message)
            self._X_train_umap = self.dim_reducer.UMAP.fit_transform(self.X_train_pca)
        return self._X_train_umap

    @property
    def X_test(self):
        if not hasattr(self, "_X_test"):
            self._X_test = self.adata[self.test_idx, self.var_df["use_genes"]].X.A
        return self._X_test

    @property
    def X_test_scaled(self):
        if not hasattr(self, "_X_test_scaled"):
            message = "Transforming raw test data using pre-fit scaling model."
            self.t_elapsed_message(message)
            self._X_test_scaled = self.dim_reducer.Scaler.transform(self.X_test)
        return self._X_test_scaled

    @property
    def X_test_pca(self):
        if not hasattr(self, "_X_test_pca"):
            message = "Transforming scaled test data using pre-fit PCA model."
            self.t_elapsed_message(message)
            self._X_test_pca = self.dim_reducer.PCA.transform(self.X_test_scaled)
        return self._X_test_pca

    @property
    def X_test_umap(self):
        if not hasattr(self, "_X_test_umap"):
            message = "Transforming PCA test data using pre-fit UMAP model."
            self.t_elapsed_message(message)
            self._X_test_umap = self.dim_reducer.UMAP.transform(self.X_test_pca)
        return self._X_test_umap

    def concat_train_test(self, adata_task):

        train_df = self.obs_df.loc[self.train_idx].copy()
        
        if len(self.test_vals) > 0:
            test_df  = self.obs_df.loc[self.test_idx].copy()
            adata_task.obs = pd.concat([train_df, test_df])
            
        else:
            adata_task.obs = train_df
            
        return adata_task[adata_task.obs.index.astype(int).argsort()].copy()

    def __call__(self):
        
        X_train_scaled = self.X_train_scaled
        X_train_pca = self.X_train_pca
        X_train_umap = self.X_train_umap
        
        if len(self.test_vals) > 0:
            
            X = np.concatenate(
                [
                    self.adata[self.train_idx, self.var_df["use_genes"]].X.A,
                    self.adata[self.test_idx, self.var_df["use_genes"]].X.A,
                ]
            )        

            X_test_scaled = self.X_test_scaled
            X_test_pca = self.X_test_pca
            X_test_umap = self.X_test_umap

            X_scaled = np.concatenate([X_train_scaled, X_test_scaled])
            X_pca = np.concatenate([X_train_pca, X_test_pca])
            X_umap = np.concatenate([X_train_umap, X_test_umap])
            
        else:
            
            X = self.adata[self.train_idx, self.var_df["use_genes"]].X.A
            X_scaled = X_train_scaled
            X_pca = X_train_pca
            X_umap = X_train_umap

        adata_task = anndata.AnnData(X, dtype=X.dtype)
        adata_task.layers["X_scaled"] = X_scaled
        adata_task.obsm["X_pca"]  = X_pca
        adata_task.obsm["X_umap"] = X_umap

        self.adata_task = self.concat_train_test(adata_task)

        return self.adata_task
    
    
def split_for_timepoint_recovery_task(adata, split_key="Time point", write_h5ad=False):

    task_split = SplitDataForTask(
        adata, split_key=split_key, train_vals=[2, 6], test_vals=[4]
    )
    adata_task = task_split()

    adata_task.obs["train"] = adata_task.obs[task_split.split_key].isin([2, 6])
    adata_task.obs["test"] = adata_task.obs[task_split.split_key].isin([2, 4])

    if write_h5ad:
        adata_task.write_h5ad(write_h5ad)

    return adata_task

def split_for_fate_prediction_task(adata, split_key="Well", write_h5ad=False):

    task_split = SplitDataForTask(
        adata, split_key=split_key, train_vals=[0, 1], test_vals=[2]
    )
    adata_task = task_split()

    adata_task.obs["train"] = adata_task.obs[task_split.split_key].isin([0, 1])
    adata_task.obs["test"] = adata_task.obs[task_split.split_key].isin([0, 2])

    if write_h5ad:
        adata_task.write_h5ad(write_h5ad)

    return adata_task


def split_for_transfer_learning_task(adata, split_key="Time point", write_h5ad=False):

    task_split = SplitDataForTask(
        adata, split_key=split_key, train_vals=[2, 4, 6], test_vals=[]
    )
    adata_task = task_split()
    adata_task.obs["train"] = adata_task.obs[task_split.split_key].isin([2, 4, 6])

    if write_h5ad:
        adata_task.write_h5ad(write_h5ad)

    return adata_task



# lt funcs init

from ._time_occupance import time_occupance
from ._cell_fates import CellFates
from ._index_subsets import IndexSubsets
from ._annotate_fated import annotate_fated
from ._count_fate_values import count_fate_values
from ._to_anndataset import to_AnnDataset

from ._fetch_fate_bias_data import fetch_fate_bias_data

from ._annotate_clone_idx_in_obs import annotate_clone_idx_in_obs

from ._subset_clonal import subset_clonal




import pandas as pd
import ABCParse

class LineageClassification(ABCParse.ABCParse):
    """Sort ground truth lineages based on their types."""

    def __init__(self, F_obs: pd.DataFrame):
        self.__parse__(locals())

    @property
    def homogeneity(self):
        return abs(self.F_obs.max(1) - self.F_obs.sum(1))

    @property
    def impure_idx_mask(self):
        return self.homogeneity > 0

    @property
    def pure_idx_mask(self):
        return self.homogeneity == 0

    @property
    def n_impure(self):
        return self.impure_idx_mask.sum()

    @property
    def n_pure(self):
        return self.pure_idx_mask.sum()

    @property
    def F_obs_pure(self):
        return self.F_obs[self.pure_idx_mask]

    @property
    def F_obs_impure(self):
        return self.F_obs[self.impure_idx_mask]



from .... import utils


import pandas as pd
import ABCParse


class MultifatedLineagePredictionResults(ABCParse.ABCParse):
    
    def __init__(self, threshold=0.3):

        """threshold: accuracy threshold for which we decide it counts as truly labeled"""

        self.__parse__(locals(), private=["threshold"])
        
        import sklearn

        self.results = {}
        self.labels = []

    def _target_label(self, row):
        return "_".join(row[row > self._threshold].sort_values(ascending=False).index)

    def _predicted_label(self, predicted):
        if predicted.sum() == 0:
            return "None"
        return "_".join(
            predicted[predicted > self._threshold].sort_values(ascending=False).index
        )

    def _compose_target_pred_df(self, idx, predicted, target):

        target_pred_df = pd.concat(
            [
                predicted.to_frame().rename({idx: "predicted"}, axis=1),
                target.to_frame().rename({idx: "target"}, axis=1),
            ],
            axis=1,
        ).fillna(0)
        return target_pred_df[target_pred_df.sum(1) > 0]

    def _mean_abs_error(self, target_pred_df):
        return sklearn.metrics.mean_absolute_error(
            target_pred_df["predicted"].values, target_pred_df["target"].values
        )

    def __call__(self, F_obs_impure, F_hat):

        for en, (idx, row) in enumerate(F_obs_impure.iterrows()):

            target = row[row > 0]
            target_label = self._target_label(row)
            predicted = F_hat.loc[idx]
            predicted_label = self._predicted_label(predicted)
            for label in [target_label, predicted_label]:
                if not label in self.labels:
                    self.labels.append(label)

            error = self._mean_abs_error(
                self._compose_target_pred_df(idx, predicted, target)
            )

            self.results[idx] = {
                "target": target_label,
                "predicted": predicted_label,
                "error": error,
            }

        self.labels = sorted(self.labels)
        self.results_df = pd.DataFrame(self.results).T

        return self.results_df



import os
from pathlib import Path
from typing import Dict
import wget
import licorice_font
import ABCParse

from .. import utils
from ._directory_manager import mkdir


# -- base classes: -------------------------------------------------------------
class FormattedURLFile(ABCParse.ABCParse):
    def __init__(self, file, download_dir, download_structure, base_url):

        self.__parse__(locals(), public=[None])
    
    @property
    def _fname(self):
        return os.path.basename(self._file)

    @property
    def URL(self):
        return os.path.join(self._base_url, self._fname)

    @property
    def _download_path(self):
        return Path(os.path.join(self._download_dir, self._download_structure)).absolute()

    @property
    def fpath(self):
        return Path(os.path.join(self._download_path, self._fname)).absolute()

    @property
    def downloaded(self):
        return self.fpath.exists()
        

class URLPathInterface:
    
    def __init__(self, download_path=os.getcwd()):
        
        self._download_path = download_path
        self._base_url = "https://kleintools.hms.harvard.edu/paper_websites/state_fate2020/"

    def _format(self, file):
        formatted = FormattedURLFile(
            file=f"stateFate_{self._dataset}_{file}",
            download_dir=self._download_path,
            download_structure=self._download_structure,
            base_url=self._base_url,
            )
        self._download_dir = formatted._download_path
        return formatted

    @property
    def gene_names(self):
        return self._format(file = "gene_names.txt.gz")

    @property
    def clone_matrix(self):
        return self._format(file = "clone_matrix.mtx.gz")

    @property
    def normed_counts(self):
        return self._format(file = "normed_counts.mtx.gz")

    @property
    def metadata(self):
        return self._format(file = "metadata.txt.gz")

    @property
    def __attributes__(self)->Dict:

        AttrDict = {}
        for attr in self.__dir__():
            if not attr.startswith("_") and attr != "download":
                AttrDict[attr] = getattr(self, attr)

        return AttrDict

    def download(self, silent = False, bar = False):

        for fname, file in self.__attributes__.items():
            if not file.downloaded:
                path = str(file.fpath)
                mkdir(path = path, silent = silent)
                wget.download(url=file.URL, out= path, bar=bar)

    @property
    def _attrs(self):
        return {
            attr: getattr(self, attr)
            for attr in self.__dir__()
            if not attr.startswith("_")
        }

    @property
    def _header(self):
        header_items = [
            "{:<13}".format("Component"),
            "{:^10}".format("Downloaded"),
            "Filepath",
        ]
        download_header = (
            " | ".join(
                [licorice_font.font_format(item, ["BOLD"]) for item in header_items]
            )
            + "\n"
            + "".join(["-"] * 38)
            + "\n"
        )
        return download_header

    @property
    def _description(self):
        header = self._header
        for k, v in self._attrs.items():
            if not k == "download":
                if v.downloaded:
                    color = "GREEN"
                else:
                    color = "RED"
                header += "{:<13} |{:^25}| {}\n".format(
                    k,
                    licorice_font.font_format(str(v.downloaded), ["BOLD", color]),
                    v.fpath,
                )
        return header
    
    def __repr__(self):
        return self._description


# -- API-facing sub-classes: -------------------------------------------------
class inVitroURLPaths(URLPathInterface):
    _download_structure = "KleinLabData/in_vitro"
    _dataset = "inVitro"
    def __init__(self, download_path=os.getcwd()):
        super(inVitroURLPaths, self).__init__(download_path=download_path)


class inVivoURLPaths(URLPathInterface):
    _download_structure = "KleinLabData/in_vivo"
    _dataset = "inVivo"
    def __init__(self, download_path=os.getcwd()):
        super(inVivoURLPaths, self).__init__(download_path=download_path)


class CytokinePerturbationURLPaths(URLPathInterface):
    _download_structure = "KleinLabData/cytokine_perturbation"
    _dataset = "cytokinePerturbation"
    def __init__(self, download_path=os.getcwd()):
        super(CytokinePerturbationURLPaths, self).__init__(download_path=download_path)
        



__module_name__ = "_estimate_growth_rates.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# import packages: ------------------------------------------------------------
import anndata
import numpy as np
# import vinplots
import matplotlib.pyplot as plt


# supporting plot function: ---------------------------------------------------
def plot_growth_rates(
    adata: anndata.AnnData,
    colors: list = ["dodgerblue", "crimson"],
    labels: list = ["d4", "d6"],
) -> None:

    """

    Parameters:
    -----------

    Returns:
    --------
    None
    """

    count_dict = adata.uns["n_d2_daughters"]
    count_idx = (count_dict[4] + count_dict[6]).argsort()
    growth_rates = adata.uns["GrowthRateDict"]

    fig, axes = vinplots.quick_plot(nplots=2, ncols=2, wspace=0.2, figsize=0.6)

    for n, d_g in enumerate([growth_rates["d2_d4"], growth_rates["d2_d6"]]):
        axes[0].scatter(
            range(len(d_g)), d_g[count_idx], s=2, c=colors[n], label=labels[n]
        )

    axes[0].legend(edgecolor="w", markerscale=2)
    axl = axes[0].set_ylabel("Relative Growth Rate", fontsize=8)
    axl = axes[0].set_xlabel("Clonal Lineage", fontsize=8)

    axes[1].set_xlabel("log(d4/d2) growth rate", fontsize=8)
    axes[1].set_ylabel("log(d6/d2) growth rate", fontsize=8)
    axes[1].scatter(growth_rates["d2_d4"], growth_rates["d2_d6"], s=2, c="k")
    st = plt.suptitle("Observed (via counting) Growth Rates", fontsize=10)


# plot growth rate UMAP functions ---------------------------------------------
def _get_plot_vmin_vmax(adata):

    growth_rates = adata.uns["GrowthRateDict"]

    all_growth_rates = np.hstack(list(growth_rates.values()))
    vmin, vmax = all_growth_rates.min(), all_growth_rates.max()

    return {"vmin": vmin, "vmax": vmax}


def _background_scatter_UMAP(
    ax, X_umap, c="lightgrey", alpha=0.2, s=1, rasterized=True
):
    ax.scatter(X_umap[:, 0], X_umap[:, 1], c=c, alpha=alpha, s=s, rasterized=rasterized)

def _d2_lineage_cell_idx(meta_df, d2_mask, t0=2, time_key="Time point"):
    return meta_df.loc[meta_df[time_key] == t0][d2_mask].index.astype(int)

def _plot_continuous_highlight_UMAP(
    ax,
    X_umap,
    subset_idx,
    c="navy",
    s=5,
    vmin=None,
    vmax=None,
    cax=None,
    cbar_shrink=0.5,
):

    if type(c) is str:
        pass
    else:
        c_idx = np.argsort(c)

    x, y = X_umap[subset_idx][c_idx, 0], X_umap[subset_idx][c_idx, 1]
    img = ax.scatter(x, y, c=c[c_idx], s=s, vmin=vmin, vmax=vmax, cmap=plt.cm.plasma)
    if cax:
        cbar = plt.colorbar(mappable=img, ax=cax, shrink=0.5)


def plot_growth_rate_UMAPs(
    adata,
    plot_keys=["d2_d4", "d2_d6"],
    titles=["log(d4/d2) observed growth rate", "log(d6/d2) observed growth rate"],
    vmin=None,
    vmax=None,
):

    if not vmin and not vmax:
        v = _get_plot_vmin_vmax(adata)

    vmin, vmax = v["vmin"], v["vmax"]
    X_umap = adata.obsm["X_umap"]
    growth_rates = adata.uns["GrowthRateDict"]

    plot_idx = _d2_lineage_cell_idx(
        adata.obs.copy(), adata.uns["d2_lin_mask"], t0=2, time_key="Time point"
    )

    fig, axes = vinplots.quick_plot(nplots=2, ncols=2, hspace=0.2, figsize_width=1.2, rm_ticks=True, spines_to_delete="all")
    
    for n, ax in enumerate(axes):
        _background_scatter_UMAP(ax, X_umap)
        _plot_continuous_highlight_UMAP(
            ax, X_umap, plot_idx, c=growth_rates[plot_keys[n]], cax=ax, vmin=vmin, vmax=vmax
        )
        ax.set_title(titles[n])

# supporting functions: -------------------------------------------------------
def _enumerate_time_pairs(t):

    """
    Get pairs of timepoints

    Parameters:
    -----------
    t
        unique time points for which there are counts
        type: list

    Returns:
    --------
    time_pairs
        type: list

    Notes:
    ------

    """

    time_pairs = []
    for i in t:
        for j in t:
            if not (i == j) and not (j < i) and not (i, j) in time_pairs:
                time_pairs.append((i, j))

    return time_pairs


def _calculate_growth_rate_from_counts(
    t0_count: np.ndarray,
    tf_count: np.ndarray,
    t0: float,
    tf: float,
    pseudocount: float = 1,
) -> np.ndarray:

    """
    Estimate growth rate from clonal lineage counts at multiple timepoints

    Parameters:
    -----------
    t0_count
        type: np.ndarray
    tf_count
        type: np.ndarray
    t0
        type: float
    tf
        type: float

    pseudocount
        type: float

    Returns:
    --------
    growth_rate
        type: np.ndarray

    Notes:
    ------
    (1) Source: https://github.com/gifford-lab/prescient-analysis/blob/master/notebooks/02b-weinreb2020-proliferation.ipynb
    """

    count_ratio = (tf_count + pseudocount) / (t0_count + pseudocount)
    return np.log(count_ratio) / (tf - t0)


# primary function: -----------------------------------------------------------
def estimate_growth_rates(
    adata: anndata.AnnData,
    pseudocount: int = 1,
    return_dict: bool = False,
    plot: bool = True,
    plot_colors: list = ["dodgerblue", "crimson"],
    plot_labels: list = ["d4", "d6"],
) -> dict:

    """
    Estimate growth rate from counts of daughter cells relative to d2 progenitors, across multiple timepoints.

    Parameters:
    -----------
    count_dict

    pseudocount

    Returns:
    --------
    GrowthRateDict
    """

    count_dict = adata.uns["n_d2_daughters"]

    GrowthRateDict = {}

    t = list(count_dict.keys())
    time_pairs = _enumerate_time_pairs(t)

    for t0, tf in time_pairs:
        key = "d{}_d{}".format(int(t0), int(tf))
        GrowthRateDict[key] = _calculate_growth_rate_from_counts(
            count_dict[t0], count_dict[tf], t0, tf, pseudocount=pseudocount
        )

    adata.uns["GrowthRateDict"] = GrowthRateDict
    
    if plot:
        plot_growth_rates(adata, colors=plot_colors, labels=plot_labels)
        plot_growth_rate_UMAPs(adata,
                               plot_keys=["d2_d4", "d2_d6"],
                               titles=["log(d4/d2) observed growth rate", "log(d6/d2) observed growth rate"],
                               vmin=None,
                               vmax=None,
                              )
    

    if return_dict:
        return GrowthRateDict



# -- import local dependencies: ------------------------------------------------
from ._figshare_downloader import FigshareDownloader


# -- import packages: ----------------------------------------------------------
import anndata
import pathlib


# -- set typing: ---------------------------------------------------------------
from typing import Optional, Union


# -- API-facing function: ------------------------------------------------------
def LARRY_timepoint_interpoltion(
    fpath: Optional[Union[str, pathlib.Path]] = None,
    url: Optional[Union[str, pathlib.Path]] = None,
) -> anndata.AnnData:
    """
    Downloader for Timepoint Recovery Task AnnData
    https://figshare.com/articles/dataset/LARRY_LT-scRNA-seq_dataset_formatted_for_timepoint_recovery/23692095
    """

    if fpath is None:
        fpath = (
            "scdiffeq_data/adata.Weinreb2020.in_vitro.task_01.timepoint_recovery.h5ad"
        )

    if url is None:
        url = "https://figshare.com/ndownloader/files/41576409"

    figshare_downloader = FigshareDownloader()
    return figshare_downloader(url=url, fpath=fpath)


def LARRY_fate_prediction(
    fpath: Optional[Union[str, pathlib.Path]] = None,
    url: Optional[Union[str, pathlib.Path]] = None,
) -> anndata.AnnData:
    """
    Downloader for Fate Prediction Task AnnData
    https://figshare.com/articles/dataset/LARRY_LT-scRNA-seq_dataset_formatted_for_fate_prediction/23699991
    """

    if fpath is None:
        fpath = (
            "scdiffeq_data/adata.Weinreb2020.in_vitro.task_02.fate_prediction.h5ad"
        )

    if url is None:
        url = "https://figshare.com/ndownloader/files/41591139"

    figshare_downloader = FigshareDownloader()
    return figshare_downloader(url=url, fpath=fpath)


def LARRY_spliced_unspliced_reads(
    fpath: Optional[Union[str, pathlib.Path]] = None,
    url: Optional[Union[str, pathlib.Path]] = None,
) -> anndata.AnnData:
    """
    Downloader for AnnData object containing spliced / unspliced reads.
    """

    if fpath is None:
        fpath = (
            "scdiffeq_data/adata.Weinreb2020.in_vitro.spliced_unspliced_reads.h5ad"
        )

    if url is None:
        url = "https://figshare.com/ndownloader/files/37028569"

    figshare_downloader = FigshareDownloader()
    return figshare_downloader(url=url, fpath=fpath)


import numpy as np


class RunningQuantile:
    def __init__(self, n_bins: int = 50):

        self.n_bins = n_bins

    def __call__(self, x, y, p):

        """calculate the quantile of y in bins of x"""

        idx = np.argsort(x)
        x = x.copy()[idx]
        y = y.copy()[idx]

        xi, xf = x[0], x[-1]
        dx = (xf - xi) / self.n_bins

        pad = dx / 2

        x_out = np.linspace(xi + pad, xf - pad, self.n_bins)
        y_out = np.zeros_like(x_out)

        for i in range(len(x_out)):

            idx = np.nonzero((x >= x_out[i] - pad) & (x < x_out[i] + pad))[0]
            if len(idx) > 0:
                y_out[i] = np.percentile(y[idx], p)
            elif i > 0:
                y_out[i] = y_out[i - 1]
            else:
                y_out[i] = np.nan

        return x_out, y_out



# -- import packages: ----------------------------------------------------------
from pathlib import Path
import anndata
import os
import scipy
import pandas as pd
import numpy as np
from licorice_font import font_format


# -- import local dependencies: ------------------------------------------------
from ._load_expr_matrix import load_expr_matrix
from ._url_path_interfaces import (
    inVitroURLPaths,
    inVivoURLPaths,
    CytokinePerturbationURLPaths,
)
from .._tools import annotate_clone_idx_in_obs, count_fate_values


from .klein_lab_pp_recipe import (
    highly_variable_genes,
    remove_cell_cycle_correlated_genes,
)

from ._split_data import (
    split_for_timepoint_recovery_task,
    split_for_fate_prediction_task,
    split_for_transfer_learning_task,
)

info = font_format("INFO", ['PURPLE'])
note = f"- [ {info} ] |"

# -- base class: ---------------------------------------------------------------
class DataHandler:
    def __init__(self, silent = False):
        # must set self._url_paths
        self.URLPaths.download()
        self._silent = silent
        
    @property
    def URLPaths(self):
        return self._url_paths

    @property
    def X(self):
        if not hasattr(self, "_X"):
            self._X = scipy.sparse.csr_matrix(load_expr_matrix(path = self.URLPaths.normed_counts.fpath))
        return self._X

    @property
    def obs(self):
        obs_df = pd.read_csv(self.URLPaths.metadata.fpath, sep="\t")
        obs_df.index = obs_df.index.astype(str)
        return obs_df

    @property
    def var(self):
        var_df = pd.read_csv(self.URLPaths.gene_names.fpath, header=None, names=["gene_ids"])
        var_df.index = var_df.index.astype(str)
        return var_df

    @property
    def X_clone(self):
        if not hasattr(self, "_X_clone"):
            self._X_clone = scipy.sparse.csr_matrix(scipy.io.mmread(self.URLPaths.clone_matrix.fpath))
            if self._X_clone.shape[0] != self.X.shape[0]:
                self._X_clone = self._X_clone.T
        return self._X_clone

    def compose_adata(self):
        self.adata = anndata.AnnData(
            X=self.X,
            dtype=self.X.dtype,
            obs=self.obs,
            var=self.var,
            obsm={"X_clone": self.X_clone},
        )
        annotate_clone_idx_in_obs(self.adata)
        if self._dataset == "in_vitro":
            count_fate_values(
                self.adata,
                origin_time=[2],
                fate_time=[4, 6],
                annotation_key='Cell type annotation',
                time_key='Time point',
                lineage_key='clone_idx',
                key_added='fate_counts',
                return_dfs=False,
            )
        return self.adata

    @property
    def raw_h5ad_path(self):
        return Path(os.path.join(self.URLPaths._download_dir, f"adata.Weinreb2020.{self._dataset}.raw.h5ad"))
    
    @property
    def gene_filtered_h5ad_path(self):
        return Path(os.path.join(self.URLPaths._download_dir, f"adata.Weinreb2020.{self._dataset}.gene_filtered.h5ad"))
    
    @property
    def timepoint_recovery_h5ad_path(self):
        return Path(os.path.join(
            self.URLPaths._download_dir,
            f"adata.Weinreb2020.{self._dataset}.task_01.timepoint_recovery.h5ad",
        ))
    
    @property
    def fate_prediction_h5ad_path(self):
        return Path(os.path.join(
            self.URLPaths._download_dir,
            f"adata.Weinreb2020.{self._dataset}.task_02.fate_prediction.h5ad",
        ))
    
    @property
    def transfer_learning_h5ad_path(self):
        return Path(os.path.join(
            self.URLPaths._download_dir,
            f"adata.Weinreb2020.{self._dataset}.task_03.transfer_learning.h5ad",
        ))

    def to_h5ad(self, h5ad_path):
        if not h5ad_path.exists():
            self.adata.write_h5ad(h5ad_path)
        
    def read_h5ad(self, h5ad_path):
        self.adata = anndata.read_h5ad(h5ad_path)
        if not self._silent:
            print(self.adata)
        return self.adata

    def __call__(self):        
        
        if self.gene_filtered_h5ad_path.exists():
            print(f"{note} Reading pre-filtered {self._dataset} adata from .h5ad")
            self.adata = self.read_h5ad(self.gene_filtered_h5ad_path)
            return self.adata

        if self.raw_h5ad_path.exists():
            print(f"{note} Reading raw {self._dataset} adata from .h5ad")
            self.adata = self.read_h5ad(self.raw_h5ad_path)
        else:
            print(f"{note} Composing LARRY {self._dataset} dataset to AnnData.")
            self.adata = self.compose_adata()
            print(f"{note} Saving raw adata to .h5ad")
            self.to_h5ad(self.raw_h5ad_path)
        
        print(f"{note} Calling highly variable genes")
        highly_variable_genes(self.adata)
        print(f"{note} Removing cell cycle correlated genes")
        remove_cell_cycle_correlated_genes(self.adata)
        print(f"{note} Saving gene-filtered adata to .h5ad")
        self.to_h5ad(self.gene_filtered_h5ad_path)
        
        return self.adata
        
# -- DataHandlers: -------------------------------------------------------------
class inVitroData(DataHandler):
    _url_paths = inVitroURLPaths()
    _dataset = "in_vitro"
    def __init__(self, silent = False):
        super(inVitroData, self).__init__(silent=silent)
        
        
    def fate_prediction(self, split_key="Well", write_h5ad=False):
        
        if self.fate_prediction_h5ad_path.exists():
            print(f"{note} Reading adata prepared for fate prediction task from .h5ad")
            return self.read_h5ad(self.fate_prediction_h5ad_path)
        
        self.adata = self.__call__()
        
        return split_for_fate_prediction_task(
                self.adata,
                split_key=split_key,
                write_h5ad=self.fate_prediction_h5ad_path,
            )

    def timepoint_recovery(self, split_key="Time point"):
        
        if self.timepoint_recovery_h5ad_path.exists():
            print(f"{note} Reading adata prepared for timepoint recovery task from .h5ad")
            return self.read_h5ad(self.timepoint_recovery_h5ad_path)
        
        self.adata = self.__call__()
        
        return split_for_timepoint_recovery_task(
                self.adata,
                split_key=split_key,
                write_h5ad=self.timepoint_recovery_h5ad_path,
            )
    
    def transfer_learning(self, split_key="Time point"):
        
        if self.transfer_learning_h5ad_path.exists():
            print(f"{note} Reading adata prepared for transfer learning task from .h5ad")
            return self.read_h5ad(self.transfer_learning_h5ad_path)
        
        self.adata = self.__call__()
        
        return split_for_transfer_learning_task(
                self.adata,
                split_key=split_key,
                write_h5ad=self.transfer_learning_h5ad_path,
            )
    
class inVivoData(DataHandler):
    _url_paths = inVivoURLPaths()
    _dataset = "in_vivo"
    def __init__(self, silent = False):
        super(inVivoData, self).__init__(silent=silent)


class CytokinePerturbationData(DataHandler):
    _url_paths = CytokinePerturbationURLPaths()
    _dataset = "cytokine_perturbation"
    def __init__(self, silent = False):
        super(CytokinePerturbationData, self).__init__(silent=silent)




# -- import packages: ----------------------------------------------------------
import ABCParse
import autodevice
import lightning
import pathlib
import pandas as pd

# -- import local dependencies: ------------------------------------------------
from .. import tasks

metrics = tasks.fate_prediction.metrics


# -- Callback: -----------------------------------------------------------------
class FatePredictionCallback(lightning.Callback, ABCParse.ABCParse):
    def __init__(
        self, adata, N: int = 2000, device=autodevice.AutoDevice(), *args, **kwargs
    ):
        self.__parse__(locals(), public=[None])

    def __repr__(self) -> str:
        return "FatePredictionCallback()"

    @property
    def F_obs(self):
        F_obs = tasks.fate_prediction.F_obs
        F_obs.index = F_obs.index.astype(str)
        return F_obs

    @property
    def t0_idx(self):
        return self.F_obs.index

    @property
    def _BASE_PATH(self):
        # assumes the first logger is the .CSVLogger
        base_path = pathlib.Path(self._log_dir).joinpath("fate_prediction_metrics")
        if not base_path.exists():
            base_path.mkdir()
        return base_path

    @property
    def _CKPT_METRICS_PATH(self):
        ckpt_metrics_path = self._BASE_PATH.joinpath(self._ckpt_name)
        if not ckpt_metrics_path.exists():
            ckpt_metrics_path.mkdir()
        return ckpt_metrics_path

    @property
    def _F_hat_unfiltered_csv_path(self):
        return self._CKPT_METRICS_PATH.joinpath("F_hat.unfiltered.csv")

    @property
    def _F_hat_processed_csv_path(self):
        return self._CKPT_METRICS_PATH.joinpath("F_hat.processed.csv")

    @property
    def _ACC_CSV_PATH(self):
        return self._CKPT_METRICS_PATH.joinpath("accuracy.csv")

    @property
    def _NEG_CE_CSV_PATH(self):
        return self._CKPT_METRICS_PATH.joinpath("neg_cross_entropy.csv")

    @property
    def _NM_CORR_CSV_PATH(self):
        return self._CKPT_METRICS_PATH.joinpath("neu_mon_corr.csv")

    def compute_fate_bias(self, pl_module):

        fate_bias = tasks.fate_prediction.FateBias()
        F_hat = fate_bias(
            adata=self._adata,
            DiffEq=pl_module.to(self._device),
            t0_idx=self.t0_idx,
            N=self._N,
        )
        F_hat.index = F_hat.index.astype(str)
        return F_hat

    def process_F_hat(self, F_hat: pd.DataFrame, index=None):

        print(f"Saving `F_hat` [ unfilt. ] to: {self._F_hat_unfiltered_csv_path}")
        F_hat.to_csv(self._F_hat_unfiltered_csv_path)

        F_hat = F_hat.drop("Undifferentiated", axis=1)
        F_hat = F_hat.div(F_hat.sum(1), axis=0)
        F_hat = F_hat.fillna(0)
        print(f"Saving `F_hat` [ processed ] to: {self._F_hat_processed_csv_path}")
        F_hat.to_csv(self._F_hat_processed_csv_path)

        if not index is None:
            F_hat.index = index
        else:
            F_hat.index = F_hat.index.astype(str)
        return F_hat

    def _accuracy(self, F_hat):

        accuracy_df = metrics.multi_idx_accuracy(self.F_obs, F_hat)
        print(f"Saving `accuracy_df` to: {self._ACC_CSV_PATH}")
        accuracy_df.to_csv(self._ACC_CSV_PATH)

    def _negative_cross_entropy(self, F_hat):

        neg_cross_entropy_df = metrics.multi_idx_negative_cross_entropy(
            self.F_obs, F_hat
        )
        print(f"Saving `neg_cross_entropy_df` to: {self._NEG_CE_CSV_PATH}")
        neg_cross_entropy_df.to_csv(self._NEG_CE_CSV_PATH)

    def _neu_mon_corr(self, F_hat):

        neu_mon_corr_df = pd.DataFrame(
            metrics.neutrophil_monocyte_correlation(self.F_obs, F_hat)
        )
        print(f"Saving `neu_mon_corr_df` to: {self._NM_CORR_CSV_PATH}")
        neu_mon_corr_df.to_csv(self._NM_CORR_CSV_PATH)

    def compute_metrics(self, F_hat):

        self._accuracy(F_hat)
        self._negative_cross_entropy(F_hat)
        self._neu_mon_corr(F_hat)

    def forward(self, pl_module, ckpt_name, log_dir):

        """ """

        self.__update__(locals())

        self.F_hat_unfiltered = self.compute_fate_bias(pl_module)
        self.F_hat = self.process_F_hat(self.F_hat_unfiltered)
        self.compute_metrics(self.F_hat)

    def on_train_end(self, trainer, pl_module):

        ckpt_name = f"on_train_end.epoch_{pl_module.current_epoch}"

        log_dir = self._log_dir = pl_module.loggers[0].log_dir

        self.forward(pl_module, ckpt_name=ckpt_name, log_dir=log_dir)

        
# -- import packages: ----------------------------------------------------------
# import autodevice
# import lightning
# import ABCParse

# -- import local dependencies: ------------------------------------------------
# from .. import utils
# from .. import tasks


# # -- Callback class: -----------------------------------------------------------
# class FatePredictionCallback(lightning.Callback, ABCParse.ABCParse):
#     def __init__(self, model):
        
#         self.__parse__(locals())
        
#         self._parse_model(model)

#     def _parse_model(self, model):
    
#         adata = model.adata
#         kNN_Graph = model.kNN_Graph
#         PCA = model.reducer.PCA




import ABCParse


from ... import utils
from . import _fate_prediction_utils as fate_utils


class F_obs(ABCParse.ABCParse):
    """Format ground truth comparison data and get labels, etc."""

    def __init__(
        self,
        adata,
        origin_time=[2],
        fate_time=[4, 6],
        annotation_key="Cell type annotation",
        time_key="Time point",
        lineage_key="clone_idx",
        key_added="fate_counts",
        return_dfs=False,
    ):

        self.__parse__(locals(), public=["adata"])

        fate_utils.count_fate_values(
            self.adata,
            origin_time=self._origin_time,
            fate_time=self._fate_time,
            annotation_key=self._annotation_key,
            time_key=self._time_key,
            lineage_key=self._lineage_key,
            key_added=self._key_added,
            return_dfs=self._return_dfs,
        )

        self._df = self.adata.obs.copy()
        
    @property
    def _df_COUNTS(self):
        """This is the F_obs DataFrame"""
        if not hasattr(self, "_fate_df_COUNTS"):
            _fate_df = (
                self.adata[self._df[self._time_key] == self._origin_time[0]]
                .obsm["cell_fate_df"]
                .fillna(0)
            )
            for key in ['undiff', "Undifferentiated", "clone_idx"]:
                if key in _fate_df.columns:
                    _fate_df = _fate_df.drop(key, axis=1)

            self._fate_df_COUNTS = _fate_df[_fate_df.sum(1) > 0]
            
        return self._fate_df_COUNTS

    @property
    def df(self):
        if not hasattr(self, "_fate_df"):
            self._fate_df = utils.sum_norm_df(self._df_COUNTS)
        return self._fate_df

    def __call__(self, return_counts = False):
        if return_counts:
            return self._df_COUNTS
        return self.df




__module_name__ = "_count_clones_by_timepoint.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# import packages: ------------------------------------------------------------
import anndata
import pandas


# -----------------------------------------------------------------------------
def count_clones_by_timepoint(
    adata: anndata.AnnData,
    lineage_key: str = "clone_idx",
    time_key: str = "Time point",
    key_added: str = "lineage_time_counts",
    return_df: bool = False,
) -> pandas.DataFrame:

    """
    Annotates adata with pandas.DataFrame: adata.uns['clone_x_timepoint'].
    
    Parameters:
    -----------
    adata
        type: anndata.AnnData
        
    lineage_key
        type: str
        default: "clone_idx"
    time_key
        type: str
        default: "Time point"
    
    return_df
        type: bool
        default: False
    
    Returns:
    --------
    None or clone_x_timepoint
        type: NoneType or pandas.DataFrame
    """

    obs_df = adata.obs.copy()

    clone_x_timepoint = (
        obs_df.dropna()
        .groupby("Time point")["clone_idx"]
        .value_counts()
        .to_frame()
        .unstack()
        .fillna(0)
    ).T
    
    clone_x_timepoint = clone_x_timepoint.reset_index(drop=True)
    clone_x_timepoint.columns.name = None

    adata.uns["clone_x_timepoint"] = clone_x_timepoint

    if return_df:
        return clone_x_timepoint


import pathlib
import pandas as pd


class _F_obs:
    def __init__(self):
        """"""

    @property
    def _PATH(self):
        return pathlib.Path(__file__).parent.joinpath("F_obs.csv")

    @property
    def df(self):

        df = pd.read_csv(self._PATH, index_col=0)
        df.index = df.index.astype(str)

        return df

    def __call__(self):
        return self.df




import pandas as pd
import ABCParse

from ._in_vitro_color_map import InVitroColorMap
from .. import utils


class MultiFateInVitroColorMap(ABCParse.ABCParse):
    def __init__(self, save=False, fname="label_cmap.csv"):

        self.__parse__(locals(), public=[None])
        self._CMAP = InVitroColorMap()._dict
        self.__build_cmap__()
        if self._save:
            self._to_csv()

    @property
    def Baso(self):
        return {
            "Baso": "#0a9396",
            "Baso_Eos": "#29a1a4",
            "Baso_Eos_Neutrophil": "#48aeb1",
            "Baso_Erythroid": "#67bcbe",
            "Baso_Mast": "#85c9cb",
            "Baso_Meg": "#a4d7d8",
            "Baso_Monocyte": "#c2e4e5",
            "Baso_Neutrophil": "#e1f2f2",
        }

    @property
    def Ccr_7(self):
        return {
            "Ccr_7": "#ae2012",
            "Ccr7_DC_Lymphoid": "#ae2012",
        }

    @property
    def Eos(self):
        return {
            "Eos": "#005F73",
            "Eos_Baso": "#80AFB9",
            "Eos_Neutrophil": "#C0D7DC",
        }

    @property
    def Erythroid(self) -> dict:
        return {
            "Erythroid": "#E9D8A6",
            "Erythroid_Mast": "#EFE2BD",
            "Erythroid_Meg": "#FAF6E9",
        }

    @property
    def Lymphoid(self) -> dict:
        return {
            "Lymphoid": "#EE9B00",
            "Lymphoid_Monocyte": "#F3B440",
            "Lymphoid_pDC": "#F7CD80",
        }

    @property
    def Mast(self) -> dict:
        return {
            "Mast": "#94D2BD",
            "Mast_Baso": "#AFDECE",
            "Mast_Erythroid": "#BDE4D6",
            "Mast_Meg": "#CAE9DE",
            "Mast_Neutrophil": "#E5F4EF",
        }

    @property
    def Meg(self) -> dict:
        return {
            "Meg": "#9B2226",
            "Meg_Baso": "#A83E42",
            "Meg_Erythroid": "#B45A5D",
            "Meg_Monocyte": "#CD9193",
            "Meg_Neutrophil": "#E6C8C9",
        }

    @property
    def Monocyte(self) -> dict:
        return {
            "Monocyte": "#F08700",
            "Monocyte_Lymphoid": "#F29620",
            "Monocyte_Mast": "#F4A540",
            "Monocyte_Neutrophil": "#F8C380",
        }

    @property
    def Neutrophil(self) -> dict:
        return {
            "Neutrophil": "#023047",
            "Neutrophil_Baso": "#224A5E",
            "Neutrophil_Ccr7_DC": "#426475",
            "Neutrophil_Monocyte": "#8198A3",
        }

    @property
    def cmap(self):
        return (
            pd.DataFrame.from_dict(self._CMAP, orient="index")
            .reset_index()
            .rename({"index": "label", 0: "color"}, axis=1)
        )

    def __build_cmap__(self):
        for attr in self.__dir__():
            if not (attr.startswith("_")) and (attr != "cmap"):
                self._CMAP.update(getattr(self, attr))

    def _to_csv(self):
        self.cmap.to_csv(self._fname)
        
def multi_fate_in_vitro_cmap():
    return MultiFateInVitroColorMap().cmap




import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import os
import ABCParse

from ._multi_fate_in_vitro_color_map import multi_fate_in_vitro_cmap

from typing import List, Optional, Tuple

class ConfusionMatrix(ABCParse.ABCParse):
    def __init__(self, *args, **kwargs) -> None:

        self.__parse__(locals())
        
        import seaborn as sns
        import sklearn
        
        self.multi_fate_in_vitro_cmap = multi_fate_in_vitro_cmap()
        
    def _configure_labels(self):
        """If no labels are passed, confiugure labels on the fly from self.F_obs"""
        return self.F_obs.sum(0).sort_values(ascending=False).index.tolist()
        
    @property
    def labels(self) -> List[str]:
        if self._labels is None:
            self._labels = self._configure_labels()
        return self._labels

    @property
    def y_true(self) -> np.ndarray:
        if isinstance(self.F_obs, pd.DataFrame):
            return self.F_obs.idxmax(1).values
        return self.F_obs

    @property
    def y_pred(self):
        if isinstance(self.F_hat, pd.DataFrame):
            return self.F_hat.idxmax(1).values
        return self.F_hat

    @property
    def conf_mtx(self):
        import sklearn
        return sklearn.metrics.confusion_matrix(
            y_true=self.y_true,
            y_pred=self.y_pred,
            labels=self.labels,
        )

    @property
    def _ANNOT(self):
        _ANNOT = self.conf_mtx.astype(str)
        _ANNOT[np.where(_ANNOT.astype(str) == "0")] = ""
        return _ANNOT
    
    def _format_max_val(self):
        max_val = self.conf_mtx.max()
        
        if max_val >= 100:
            divisor, rounder = 100, -2
        else:
            divisor, rounder = 10, -1
        
        diff = divisor - (max_val % divisor)
        return np.round(max_val + diff, rounder)

    @property
    def _VMAX(self):
        if self._vmax is None:
            return self._format_max_val()
        return self._vmax
    
    @property
    def _CMAP(self):
        cmap_labels = []
        _df = self.multi_fate_in_vitro_cmap.copy()
        return [
            _df.loc[_df["label"] == label]["color"].values[0] for label in self.labels
        ]

    def __heatmap__(self):
        
        import seaborn as sns
        
        cg = sns.clustermap(
            self.conf_mtx,
            figsize=self._figsize,
            cmap=self._cmap,
            annot=self._ANNOT,
            xticklabels=self.labels,
            yticklabels=self.labels,
            row_colors=self._CMAP,
            col_colors=self._CMAP,
            row_cluster=False,
            col_cluster=False,
            fmt="",
            vmin=0,
            vmax=self._VMAX,
            cbar_pos=(0.92, 0.30, 0.02, 0.4),
            annot_kws={"size": 6},
        )
        cg.ax_heatmap.tick_params(axis="both", which="both", labelsize=6)
        cg.ax_heatmap.set_title(
            self._title,
            fontsize = self._title_fontsize,
            y = 1.04,
        )

    def __call__(
        self,
        F_obs,
        F_hat,
        labels=None,
        title=None,
        title_fontsize = 10,
        ax=None,
        cmap="Blues",
        vmax=None,
        figsize=(4, 4),
        save=False,
        *args,
        **kwargs,
    ):

        self.__parse__(locals(), public = ["F_obs", "F_hat"])
        
        self.__heatmap__()

        if save:
            plt.savefig(save)
            

def confusion_matrix(
    F_obs: pd.DataFrame,
    F_hat: pd.DataFrame,
    labels: Optional[List[str]] = None,
    title: Optional[str] = None,
    ax: Optional[plt.Axes] = None,
    cmap="Blues",
    vmax=None,
    figsize: Tuple[float, float] = (4, 4),
    save=False,
    *args,
    **kwargs,
):
    """
    Compute and plot confusion matrix between observed (`F_obs`) and predicted (`F_hat`).
    
    Parameters:
    -----------
    Returns:
    --------
    
    """
    
    confusion_mtx = ConfusionMatrix()
    KWARGS = ABCParse.function_kwargs(func = confusion_mtx, kwargs = locals())
    return confusion_mtx(**KWARGS)




# -- import packages: ---------------
import ABCParse
from typing import List
import scipy.stats
import pandas as pd

# -- import local dependencies: -----
from .._in_vitro_fate_prediction_subsets import InVitroFatePredictionSubsets

# -- operational class: -------------
class NeuMonCorr(ABCParse.ABCParse):
    def __init__(
        self,
        fate: str = "Neutrophil",
        possible_fates: List[str] = ["Neutrophil", "Monocyte"],
        corr_func=scipy.stats.pearsonr,
        *args,
        **kwargs,
    ):
        """"""
        self.__parse__(locals())
        
        assert fate in possible_fates
        self._indices = InVitroFatePredictionSubsets().indices
        
    @property
    def train_early_nm_idx(self):
        return self._indices['unique_train']['N/M-early']
        
    @property
    def test_early_nm_idx(self):
        return self._indices['unique_test']['N/M-early']
        
    @property
    def train_nm_idx(self):
        return self._indices['unique_train']['N/M']
        
    @property
    def test_nm_idx(self):
        return self._indices['unique_test']['N/M']
        
    @property
    def n_fates(self):
        return len(self._possible_fates)
    
    @property
    def na_fill(self):
        """What nan values should be filled with, effectively substituting a +1 rule"""
        return 1 / self.n_fates
        
    def row_normalize(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Sum normalize a pandas DataFrame across rows.

        Parameters
        ----------
        df: pd.DataFrame
            Input (raw, un-normalized) df

        Returns
        -------
        df_norm: pd.DataFrame
            Row-sum normalized df
        """
        return df.div(df.sum(axis=1), axis=0)
    
    @property
    def true_fate_bias(self):
        if not hasattr(self, "_true_fate_bias"):
            self._true_fate_bias = self.row_normalize(self._F_obs[self._possible_fates]).fillna(self.na_fill)[self._fate]
        return self._true_fate_bias
        
    @property
    def pred_fate_bias(self):
        if not hasattr(self, "_pred_fate_bias"):
            self._pred_fate_bias = self.row_normalize(self._F_hat[self._possible_fates]).fillna(self.na_fill)[self._fate]
        return self._pred_fate_bias
    
    def correlation(self, index):
        
        F_obs = self.true_fate_bias
        F_hat = self.pred_fate_bias
        
        return self._corr_func(F_obs.loc[index], F_hat.loc[index])
    
    @property
    def _ATTRS(self):
        return [key for key in self.__dir__() if "nm_idx" in key]
    
    def _to_dict(self, result):
        return {
            "correlation": result.correlation,
            "statistic": result.statistic,
            "pvalue": result.pvalue,
            "CI_low": result.confidence_interval().low,
            "CI_high": result.confidence_interval().high,
        }

    def forward(self, attr):
        ix = getattr(self, attr)
        return self._to_dict(self.correlation(ix))
    
    def _to_frame(self, Results):
        return pd.DataFrame(Results).T
    
    def __call__(
        self,
        F_obs: pd.DataFrame,
        F_hat: pd.DataFrame,
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        
        """
        Args:
            F_obs (pd.DataFrame): true fate bias matrix.

            F_hat (pd.DataFrame): true fate bias matrix.
            
        Returns:
            (pd.DataFrame)
        """

        self.__parse__(locals())
        
        Results = {}
        for attr in self._ATTRS:
            attr_key = attr.split("_idx")[0]
            Results[attr_key] = self.forward(attr)
        return self._to_frame(Results)
    
    

# -- API-facing function: --------------------------------------------            
def neutrophil_monocyte_correlation(
    F_obs: pd.DataFrame,
    F_hat: pd.DataFrame,
    *args,
    **kwargs,
) -> pd.DataFrame:
    """
    Args:
        F_obs (pd.DataFrame): true fate bias matrix.
        
        F_hat (pd.DataFrame): true fate bias matrix.
    """
    nm_corr = NeuMonCorr()
    return nm_corr(F_obs, F_hat)


# import ABCParse
# from typing import List
# import scipy.stats

# from .... import utils

# class NeuMonCorr(ABCParse.ABCParse):
#     def __init__(
#         self,
#         key: str = "Neutrophil",
#         reference: List[str] = ["Neutrophil", "Monocyte"],
#         corr_func=scipy.stats.pearsonr,
#         *args,
#         **kwargs,
#     ):
#         self.__parse__(locals())

#     def forward(self, df):
#         return utils.row_norm_df(df[self._reference]).fillna(
#             1 / len(self._reference)
#         )[self._key]

#     @property
#     def correlation(self):
#         return self._corr_func(self._F_obs_, self._F_hat_)

#     def __call__(self, F_obs, F_hat, *args, **kwargs):

#         self.__parse__(locals())

#         self._F_obs_ = self.forward(self._F_obs)
#         self._F_hat_ = self.forward(self._F_hat)

#         return self.correlation
    
# def neutrophil_monocyte_correlation(F_obs, F_hat):
#     nm_corr = NeuMonCorr()
#     return nm_corr(F_obs, F_hat)




# -- import packages: ----------------------------------------------------------
import licorice_font


# -- API-interacting class: ----------------------------------------------------
class Messages:
    """Container for formatted messages"""

    def __init__(self):

        self._info = licorice_font.font_format("INFO", ["PURPLE"])

    @property
    def note(self):
        """Returns:' - [ INFO ] |'"""
        return f"- [ {self._info} ] |"
    
    def mtx_to_npz(self):
        self._mtx_to_npz = "Loading expression matrix as `.mtx` (slow). Will save as `.npz` for much faster future loading."
        msg = self.note + self._mtx_to_npz
        print(msg)
    
    def reading_raw_h5ad(self, dataset):
        print(self.note + f"Reading raw {dataset} adata from .h5ad")
        
    def reading_gene_filtered_h5ad(self, dataset):
        print(self.note + f"Reading gene-filtered {dataset} adata from .h5ad")
        
    def reading_timepoint_recovery_preprocessed_h5ad(self, dataset):
        print(self.note + f"Reading {dataset} adata preprocessed for timepoint recovery from .h5ad")
        
    def reading_fate_prediction_preprocessed_h5ad(self, dataset):
        print(self.note + f"Reading {dataset} adata preprocessed for fate prediction from .h5ad")
        
    def reading_uniformly_preprocessed_h5ad(self, dataset):
        print(self.note + f"Reading uniformly preprocessed {dataset} adata from .h5ad")


import requests
import anndata
import pathlib
import os

from ..utils import InfoMessage

class FigshareDownloader:
    """Download AnnData stored in Figshare"""

    def __init__(
        self, block_size: int = 1024, notebook: bool = True, silent: bool 
= False
    ):
        """
        # 1 Kibibyte
        """
        self._block_size = block_size
        self._notebook = notebook
        self._INFO = InfoMessage()

    @property
    def _RESPONSE(self):
        if not hasattr(self, "_response"):
            self._response = requests.get(self._url, stream=True)
            self._response.raise_for_status()
        return self._response

    @property
    def _TOTAL(self):
        """"""
        return int(self._RESPONSE.headers.get("content-length", 0))

    @property
    def _TQDM(self):
        if not hasattr(self, "_tqdm"):
            if self._notebook:
                import tqdm.notebook

                tqdm = tqdm.notebook.tqdm
            else:
                from tqdm import tqdm

            self._tqdm = tqdm
        return self._tqdm

    @property
    def _PROGRESS_BAR(self):
        if not hasattr(self, "_progress_bar"):
            self._progress_bar = self._TQDM(
                total=self._TOTAL, unit="iB", unit_scale=True
            )
        return self._progress_bar

    @property
    def _PASSES_DOWNLOAD_QC_CHECK(self):
        return self._PROGRESS_BAR.n == self._TOTAL

    @property
    def fpath(self):
        return pathlib.Path(self._fpath).absolute()

    @property
    def _FILE_DIR(self):
        return self.fpath.parent

    def _configure_file_destination(self):
        if not self._FILE_DIR.exists():
            self._FILE_DIR.mkdir()

    @property
    def _FILE_DOWNLOADED(self):
        return os.path.exists(self._fpath)

    def _read_adata(self):
        if not hasattr(self, "_adata"):
            self._adata = anndata.read_h5ad(self._fpath)
            self._adata.uns["source_url"] = self._url
        return self._adata

    def __call__(self, url, fpath, return_adata=True):
        """


        Parameters
        ----------

        Returns
        -------
        """

        self._url = url
        self._fpath = fpath

        self._configure_file_destination()

        if not self._FILE_DOWNLOADED:
            self._INFO(f"Downloading to: {fpath}")
            with open(self.fpath, "wb") as file:
                for data in self._RESPONSE.iter_content(self._block_size):
                    self._PROGRESS_BAR.update(len(data))
                    file.write(data)
            self._PROGRESS_BAR.close()
            self._PASSES_DOWNLOAD_QC_CHECK

        if return_adata:
            return self._read_adata()
        return self.fpath




from ... import utils
from . import _fate_prediction_utils as fate_utils

import pandas as pd
import autodevice
import torch
import time
import ABCParse
import adata_query

# from ._fate_prediction_data import FatePredictionData
from ._observed_fate_bias import F_obs

from typing import List, Optional, Union


### from when the `FatePredictionData` class was in it's own module script: 
### MOVE THIS TO _fate_bias_matrix_generation.py ?? and make one class. would make sense.
### then you have the task class which encompasses the modules for accuracy quantitation.
def is_notebook() -> bool:
    """
    https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook#:~:text=def%20is_notebook()%20%2D%3E,standard%20Python%20interpreter
    """
    
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter
    

class FatePredictionData(ABCParse.ABCParse):
    def __init__(
        self,
        adata,
        time_key = "Time point",
        use_key = "X_pca",
        N = 2000,
        t0_idx: Optional[Union[List, pd.Index]] = None,
        groupby: Optional[Union[None, str]] = None,
        device: str = autodevice.AutoDevice(),
    ):
        
        self.__parse__(locals(), public = ['adata'])
        self.df = adata.obs.copy()
        
    @property
    def _t(self):
        if not hasattr(self, "_t_list"):
            self._t_list = sorted(self.adata.obs[self._time_key].unique().tolist())
        return self._t_list
        
    @property
    def t(self):
        return torch.Tensor(self._t).to(self._device)
        
    @property
    def t0_idx(self):
        if not hasattr(self, "_t0_idx"):
            self._t0_idx = self.df.loc[self.df[self._time_key] == self._t[0]].index
        return self._t0_idx
    
    @property
    def t0_adata(self):
        if not hasattr(self, "_t0_adata"):
            self._t0_adata = self.adata[self.t0_idx]
        return self._t0_adata
        
    @property
    def X0(self):
        if not hasattr(self, "_X0"):
            
            self._X0 = adata_query.fetch(
                adata = self.t0_adata,
                key = self._use_key,
                torch = True,
                groupby = None,
                device = autodevice.AutoDevice(),
            )[:, None, :].expand(-1, self._N, -1)
            
#             self._X0 = fate_utils.fetch_data(
#                 adata = self.t0_adata,
#                 use_key=self._use_key,
#                 torch = True,
#                 groupby = None,
#                 device = autodevice.AutoDevice(),
#             )[:, None, :].expand(-1, self._N, -1)

        return self._X0


# -- main class: ----------------------------------------------------------------------      
class FateBias(ABCParse.ABCParse):
    def __init__(self, device=autodevice.AutoDevice()):
        
        self.__parse__(locals())
        
    @property
    def kNN(self):
        if not hasattr(self, "_kNN"):
            import scdiffeq as sdq
            self._kNN = sdq.tl.kNN(self._adata)
        return self._kNN
    
    @property
    def _MODEL_TYPE(self):
        return str(self._DiffEq)
    
    def _fate_bias_matrix(self):
        value_counts = {
            key: val[self._obs_key].value_counts() for key, val in self.F_hat.items()
        }
        # don't divide by N: integer counts are sometimes useful
        return pd.DataFrame(value_counts).T.fillna(0) #  / self._N

    def _predict_state(self, X0, t0_idx):
        """returns only the FINAL state"""
        return self._DiffEq(X0, t=self.data.t)[-1].detach().cpu().numpy()
    
    def umap_transform(self, umap_model, t0_idx=[], random=0):

        if random > 0:
            t0_idx = np.random.choice(self.t0_idx, random)

        self.X_umap = {}

        for i in tqdm.notebook.tqdm(range(len(t0_idx)), desc="umap projection"):
            self.X_umap[t0_idx[i]] = np.stack(
                [
                    umap_model.transform(x)
                    for x in sdq.tl.DataFormat(self.X_hat[t0_idx[i]]).to_numpy()
                ]
            )
    @property
    def _kNN_BASIS(self):
        return self.kNN.X_use.shape[1]
            
    def _configure_dimension_reduction(self)->bool:
        self._dimension_reduce = self.data.X0.shape[-1] > self._kNN_BASIS
    
    @property
    def _DIMENSION_REDUCE(self)->bool:
        if not hasattr(self, "_dimension_reduce"):
            self._configure_dimension_reduction()
        return self._dimension_reduce
    
    def _configure_t0_idx(self):
        
        if hasattr(self, "_t0_idx"):
            self._t0_idx_config = self._t0_idx
        else:
            try:
                self.F_obs = F_obs
            except:
                self.F_obs = F_obs(self.adata)
                self._t0_idx_config = self.F_obs().index
            finally:
                self._t0_idx_config = self.data.t0_idx
        
    @property
    def t0_idx(self):
        if not hasattr(self, "_t0_idx_config"):
            self._configure_t0_idx()
        return self._t0_idx_config
        
    def __call__(
        self,
        adata,
        DiffEq,
        obs_key: str = "Cell type annotation",
        use_key: str = "X_pca",
        time_key: str = "Time point",
        t0_idx = None,
        N=2000,
        kNN = None,
        PCA = None,
        notebook: bool = is_notebook(),
    ):
        """
        
        Generates F from simulations of input data and a model
        
        DiffEq and Graph are required.
        """

        self.__update__(locals())
        
        self.F_hat = {}
        
        self.data = FatePredictionData(
            self._adata,
            t0_idx = self._t0_idx,
            time_key = self._time_key,
            use_key = self._use_key,
            N = self._N,
            device = self._device,
        )
        
        if notebook:
            import tqdm.notebook
            TQDM = tqdm.notebook.tqdm
        else:
            import tqdm
            TQDM = tqdm.tqdm
            
        
        for i in TQDM(range(len(self.t0_idx)), desc="EVALUATION"):

            X_hat = self._predict_state(X0=self.data.X0[i].to(self._device), t0_idx=self.t0_idx[i])
            
            if self._DIMENSION_REDUCE:
                X_hat = self.PCA.transform(X_hat)
                
            self.F_hat[self.t0_idx[i]] = self.kNN.aggregate(
                X_hat, obs_key=self._obs_key, max_only=True
            )
            
        return self._fate_bias_matrix()
        

    def __repr__(self):
        return f"Fate prediction evaluation | evaluating: {self._MODEL_TYPE}"





import pandas as pd


class InVitroColorMap:
    def __init__(self):
        self._dict = {
            "Neutrophil": "#023047",
            "Eos": "#005f73",
            "Baso": "#0a9396",
            "Mast": "#94d2bd",
            "Erythroid": "#e9d8a6",
            "Lymphoid": "#ee9b00",
            "Monocyte": "#F08700",
            "pDC": "#bb3e03",
            "Ccr7_DC": "#ae2012",
            "Meg": "#9b2226",
            "Undifferentiated": "#f0efeb",
            "undiff": "#f0efeb",
        }

    @property
    def _KEYS(self):
        return list(self._dict.keys())

    @property
    def _VALS(self):
        return list(self._dict.values())

    @property
    def df(self):
        return pd.DataFrame({"Cell type annotation": self._KEYS, "color": self._VALS})
    
def in_vitro_cmap():
    return InVitroColorMap().df



# -- import packages: ---------------------------------------------------------
import ABCParse
import autodevice
import lightning
import sklearn.decomposition

# -- import local dependencies: -----------------------------------------------
from .. import tasks

# -- set type hints: ----------------------------------------------------------
from typing import Optional

# -- callback cls: ------------------------------------------------------------
class InterpolationCallback(lightning.Callback, ABCParse.ABCParse):

    def __init__(
        self,
        model,
        n_samples=10_000,
        lineage_key="clone_idx",
        device=autodevice.AutoDevice(),
        backend="auto",
        PCA: Optional[sklearn.decomposition._pca.PCA] = None,  # for VAE
        silent=False,
        *args,
        **kwargs,
    ) -> None:

        adata = model.adata
        use_key = model._use_key
        time_key = model.t_config.time_key
        t0 = model.t[0].item()
        # PCA = model.reducer.PCA

        self.__parse__(locals())

        self.interpolation_task = tasks.interpolation.InterpolationTask(
            **self._TASK_KWARGS
        )

    def __repr__(self) -> str:
        return "InterpolationCallback()"

    @property
    def _TASK_KWARGS(self):
        return ABCParse.function_kwargs(
            func=tasks.interpolation.InterpolationTask,
            kwargs=self._PARAMS,
        )

    def on_train_epoch_end(self, trainer, DiffEq, *args, **kwargs) -> None:
        """"""
        d4_loss, d6_loss = self.interpolation_task(trainer, DiffEq, *args, **kwargs)

        self.log("interpolation_eval_d4_loss", d4_loss)
        self.log("interpolation_eval_d6_loss", d6_loss)




# -- import packages: ---------------------------------------------------------
import ABCParse
import adata_query
import anndata
import autodevice
import torch


# -- Operational class: -------------------------------------------------------
class InterpolationData(ABCParse.ABCParse):
    def __init__(
        self,
        adata: anndata.AnnData,
        time_key: str = "Time point",
        use_key: str = "X_pca",
        n_samples: int = 10_000,
        lineage_key: str = "clone_idx",
        device: torch.device = autodevice.AutoDevice(),
    ):
        self.__parse__(locals())

        self._configure_data()

    @property
    def t(self):
        return torch.Tensor(self._t).to(self._device)

    @property
    def df_clonal(self):
        return self.df.loc[self.df[self._lineage_key].notna()]

    @property
    def t0_idx(self):
        return self.df_clonal_d2.sample(self._n_samples, replace=True).index

    @property
    def X0(self):
        return adata_query.fetch(
            self._adata[self.t0_idx], key=self._use_key, torch=True, device=self._device
        )

    def _sample_at_t(self, t):
        return self.df_clonal.loc[self.df_clonal[self._time_key] == t]

    def set_clonal(self):
        for t in self._t:
            clonal_df = self._sample_at_t(t)
            setattr(self, f"df_clonal_d{int(t)}", clonal_df)

    def _configure_data(self):
        """ """
        
        self.df = self._adata.obs.copy()
        self._t = sorted(self.df[self._time_key].unique())
        self.set_clonal()

    @property
    def X_test_d4(self):
        """test: X_d4"""
        if not hasattr(self, "_d4_test"):
            self._d4_test = adata_query.fetch(
                self._adata[self.df_clonal_d4.index],
                key="X_pca", # always pca for d4, d6
                torch=True,
                device=self._device,
            )
        return self._d4_test

    @property
    def X_train_d6(self):
        """train: X_d6"""
        if not hasattr(self, "_d6_train"):
            self._d6_train = adata_query.fetch(
                self._adata[self.df_clonal_d6.index],
                key="X_pca", # always pca for d4, d6
                torch=True,
                device=self._device,
            )
        return self._d6_train




# -- import packages: ---------------------------------------------------------
import pandas as pd
import numpy as np
import ABCParse


# -- API-facing class: --------------------------------------------------------
class NegativeCrossEntropy(ABCParse.ABCParse):
    def __init__(self, epsilon=1e-7):

        self.__parse__(locals(), public=[None])

    def _clip_predictions(self, pred: np.ndarray) -> np.ndarray:
        """Given a defined epsilon parameter, clip predictions"""
        return np.clip(pred, self._epsilon, 1 - self._epsilon)

    def _log_transformed_predictions(self, obs: pd.DataFrame, pred_clipped: np.ndarray):
        return obs * np.log(pred_clipped)

    def _summing(self, pred_logged):
        return -np.sum(np.sum(pred_logged, axis=1))

    def __call__(self, obs: pd.DataFrame, pred: pd.DataFrame) -> pd.DataFrame:
        pred_clipped = self._clip_predictions(pred)
        pred_logged = self._log_transformed_predictions(obs, pred_clipped)
        return self._summing(pred_logged)

    def __repr__(self) -> str:
        return """NegativeCrossEntropy()"""
    





from ._fate_bias_matrix_generation import FateBias
from ._compute_fate_bias import compute_fate_bias
from ._observed_fate_bias import F_obs
from . import metrics
from ._in_vitro_fate_prediction_subsets import InVitroFatePredictionSubsets

from ._weinreb_test_set import weinreb_test_set

# -- quick F_obs df from cached .csv: ----
from ._cached_f_obs_frame import _F_obs

F_obs = _F_obs()()


from . import _analysis as an
from . import _plotting as pl
from . import _tools as tl
from . import callbacks
from . import datasets
from . import utils
from . import tasks


__all__ = ["an", "pl", "tl", "callbacks", "datasets", "utils", "tasks"]



from ._sparse_mtx_operations import (
    sparse_var,
    sparse_rowwise_multiply,
    mean_center,
    normalize_variance,
    sparse_zscore,
)

from ._info_message import InfoMessage
from ._noise import Noise
from ._sum_normalize import row_norm_df, col_norm_df



# -- import packages: -----------------------------------------------------------
from tqdm.notebook import tqdm
import numpy as np
import scipy


# -- import local dependencies: --------------------------------------------------
from ._cell_cycle_genes import cell_cycle_genes
from ... import utils


def remove_cell_cycle_correlated_genes(adata, min_corr=0.1, key_added = "use_genes"):

    """
    Remove signature-correlated genes from a list of test genes
    
    Arguments:
    ----------
    E: scipy.sparse.csc_matrix, shape (n_cells, n_genes)
        - full counts matrix
    gene_list: numpy array, shape (n_genes,)
        - full gene list
    exclude_corr_genes_list: list of list(s)
        - Each sublist is used to build a signature. Test genes correlated
          with this signature will be removed
    test_gene_idx: 1-D numpy array
        - indices of genes to test for correlation with the 
          gene signatures from exclude_corr_genes_list
    min_corr: float (default=0.1)
        - Test genes with a Pearson correlation of min_corr or higher 
          with any of the gene sets from exclude_corr_genes_list will
          be excluded

    Returns:
    --------
        numpy array of gene indices (subset of test_gene_idx) that 
        are not correlated with any of the gene signatures
        
    Source:
    https://github.com/AllonKleinLab/SPRING_dev/blob/aa52c405b6f15efd53c66f6856799dfe46e72d01/data_prep/spring_helper.py#L307-L328
    """
    
    E = adata.X.tocsc()
    gene_list = adata.var["gene_ids"].tolist()
    test_gene_idx = adata.var.loc[adata.var["hv_genes"]].index.astype(int).tolist()
    exclude_corr_genes_list = cell_cycle_genes()

    seed_ix_list = []
    for l in exclude_corr_genes_list:
        seed_ix_list.append(
            np.array([i for i in range(len(gene_list)) if gene_list[i] in l], dtype=int)
        )

    exclude_ix = []
    for iSet in tqdm(range(len(seed_ix_list))):
        seed_ix = seed_ix_list[iSet][
            E[:, seed_ix_list[iSet]].sum(axis=0).A.squeeze() > 0
        ]
        tmp = utils.sparse_zscore(E[:, seed_ix.flatten()])
        tmp = tmp.sum(1).A.squeeze()

        c = np.zeros(len(test_gene_idx))
        for iG in range(len(c)):
            c[iG], _ = scipy.stats.pearsonr(tmp, E[:, test_gene_idx[iG]].A.squeeze())

        exclude_ix.extend(
            [test_gene_idx[i] for i in range(len(test_gene_idx)) if (c[i]) >= min_corr]
        )
    exclude_ix = np.array(exclude_ix)
    filtered_idx = np.array(
        [g for g in test_gene_idx if g not in exclude_ix], dtype=int
    )
    adata.var[key_added] = adata.var.index.astype(int).isin(filtered_idx)



# -- import packages: ---------------------------------------------------------
import sklearn.metrics
import pandas as pd
import licorice_font
import ABCParse
import abc


# -- set typing: --------------------------------------------------------------
from typing import List


# -- import local dependencies: -----------------------------------------------
from .._in_vitro_fate_prediction_subsets import InVitroFatePredictionSubsets
from ._negative_cross_entropy import NegativeCrossEntropy


class AssertEqualLength(ABCParse.ABCParse):
    def __init__(self, *args, **kwargs):

        self.__parse__(locals())

    def _bold(self, msg):
        return licorice_font.font_format(msg, ["BOLD"])

    @property
    def condition(self):
        return len(self._F_obs) == len(self._F_hat)

    @property
    def error_message(self):
        fo = self._bold("F_obs")
        fh = self._bold("F_hat")
        return f"{fo} and {fh} must be of equal length"

    def __call__(self, F_obs, F_hat, *args, **kwargs):

        self.__update__(locals(), public=[None])

        assert self.condition, self.error_message


class MultiIndexScoring(ABCParse.ABCParse):
    """Calculate a metric across multiple subsets of passed indices"""

    def __init__(
        self,
        partition_keys: List[str] = ["unique_train", "unique_test"],
        fate_keys: List[str] = ["N/M", "N/M-early", "all_fates"],
        *args,
        **kwargs,
    ):
        self.__parse__(locals(), public=[None])

        self._length_check = AssertEqualLength()
        self._RESULTS_DICT = {}

    @property
    def fate_prediction_subsets(self):
        if not hasattr(self, "_subset_dict"):
            self._subset_indices = (
                InVitroFatePredictionSubsets().indices
            )
        return self._subset_indices

    def _subset(self, F_obs, F_hat, subset_idx):
        if not subset_idx is None:
            F_obs, F_hat = F_obs.loc[subset_idx], F_hat.loc[subset_idx]
        return F_obs, F_hat

    @abc.abstractmethod
    def forward(self):
        ...

    @property
    @abc.abstractmethod
    def _NAME(self):
        return

    def _iterate_over_subsets(self):

        self._RESULTS_DICT["all"] = self.forward(self._F_obs, self._F_hat)

        for pkey in self._partition_keys:
            for fkey in self._fate_keys:
                subset_idx = self.fate_prediction_subsets[pkey][fkey]
                self._RESULTS_DICT[f"{pkey}.{fkey}"] = self.forward(
                    self._F_obs, self._F_hat, subset_idx
                )

    def _to_frame(self):

        return pd.DataFrame.from_dict(self._RESULTS_DICT, orient="index").rename(
            {0: self._NAME}, axis=1
        )

    def __call__(self, F_obs, F_hat, *args, **kwargs):

        self._length_check(F_obs, F_hat)

        self.__update__(locals(), public=[None])
        self._iterate_over_subsets()

        return self._to_frame()


class MultiIndexAccuracy(MultiIndexScoring):
    def __init__(self, *args, **kwargs):
        super().__init__()

    @property
    def _NAME(self):
        return "accuracy"

    def forward(self, F_obs, F_hat, subset_idx=None):

        F_obs, F_hat = self._subset(F_obs=F_obs, F_hat=F_hat, subset_idx=subset_idx)

        return sklearn.metrics.accuracy_score(
            F_obs.idxmax(1).tolist(), F_hat.idxmax(1).tolist()
        )

    def __repr__(self) -> str:
        return """MultiIndexAccuracy()"""





class MultiIndexNegativeCrossEntropy(MultiIndexScoring):
    def __init__(self, *args, **kwargs):
        super().__init__()

        self.NegativeCrossEntropy = NegativeCrossEntropy()

    @property
    def _NAME(self):
        return "nce"

    def forward(self, F_obs, F_hat, subset_idx=None):

        F_obs, F_hat = self._subset(F_obs=F_obs, F_hat=F_hat, subset_idx=subset_idx)

        return self.NegativeCrossEntropy(F_obs, F_hat)

    def __repr__(self) -> str:
        return """MultiIndexAccuracy()"""
    

# -- API-facing functions: ----------------------------------------------------
def multi_idx_accuracy(F_obs, F_hat):

    multi_idx_acc = MultiIndexAccuracy()
    return multi_idx_acc(F_obs=F_obs, F_hat=F_hat)

def multi_idx_negative_cross_entropy(F_obs, F_hat):

    """
    
    """

    multi_idx_nce = MultiIndexNegativeCrossEntropy()
    return multi_idx_nce(F_obs=F_obs, F_hat=F_hat)




__module_name__ = "_get_lineage_obs.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# import packages: ------------------------------------------------------------
import pandas
import anndata


# -----------------------------------------------------------------------------
def get_lineage_obs(
    adata: anndata.AnnData,
    lineage_idx: float,
    lineage_key: str = "clone_idx",
    obs_key: str = "Cell type annotation",
    time_key: str = "Time point",
) -> pandas.DataFrame:

    """
    Parameters:
    -----------
    adata
        type: anndata.AnnData

    lineage_idx
        type: float

    lineage_key
        type: str
        default: "clone_idx"

    obs_key
        type: str
        default: "Cell type annotation"

    time_key
        type: str
        default: "Time point"

    Returns:
    --------
    lineage_obs
        type: pandas.DataFrame
    """

    obs_df = adata.obs.copy()

    return (
        obs_df.loc[obs_df[lineage_key] == lineage_idx]
        .groupby(time_key)[obs_key]
        .value_counts()
        .unstack()
    )


__module_name__ = "_violin_plot.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu"])


# import packages: ------------------------------------------------------------
import matplotlib.pyplot as plt
# import vinplots
import numpy as np
import anndata


# data prep function: ---------------------------------------------------------
def prepare_data_for_violin_plot(
    adata: anndata.AnnData, groupby: str = "d4_d6_major_fate_annot"
) -> dict:

    """
    Prep data for the cell fate violin plot.

    Parameters:
    -----------
    adata
        type: anndata.AnnData

    groupby
        type: str
        default: "d4_d6_major_fate_annot"

    Returns:
    --------
    ViolinPlotDataDict
        type: dict
    """

    df = adata.uns["d2_lineage_cells_metadata"]
    data_keys = list(adata.uns["GrowthRateDict"].keys())

    ViolinPlotDataDict = {}
    for key in data_keys:
        ViolinPlotDataDict[key] = {}
        for group, group_df in df.groupby(groupby):
            ViolinPlotDataDict[key][group] = group_df[key].values

    return ViolinPlotDataDict


# ViolinPlot class: -----------------------------------------------------------
class ViolinPlot(vinplots.Plot):
    def __init__(self, nplots=3, ncols=3, figsize=1.2, kwargs={}):

        self._nplots = nplots
        self._ncols = ncols
        self._figsize = figsize
        self._size_major = 6
        self._size_minor = 6
        self._rm_xticks = True
        self._rm_yticks = False

        self._mk_plot(**kwargs)

    def _format_ticks(
        self, ax, size_major=6, size_minor=6, rm_xticks=True, rm_yticks=False
    ):

        ax.tick_params(axis="both", which="major", labelsize=size_major)
        ax.tick_params(axis="both", which="minor", labelsize=size_minor)
        if rm_xticks:
            ax.set_xticks([])
        if rm_yticks:
            ax.set_yticks([])

    def _mk_plot(self, **kwargs):

        self.construct(
            nplots=self._nplots, ncols=self._ncols, figsize=self._figsize, **kwargs
        )
        self.modify_spines(ax="all", spines_to_delete=["top", "right", "bottom"])
        self.axes = self.linearize()
        for ax in self.axes:
            self._format_ticks(
                ax,
                size_major=self._size_major,
                size_minor=self._size_minor,
                rm_xticks=self._rm_xticks,
                rm_yticks=self._rm_yticks,
            )


# plot supporting functions class: --------------------------------------------
def _set_violinplot_line_colors(violin_plot, colors=None):

    line_keys = ["cmaxes", "cmins", "cmeans", "cbars"]
    for key in line_keys:
        if key in violin_plot.keys():
            c_colors = violin_plot[key].get_color()
            if not colors:
                colors = np.full(len(c_colors.flatten()), "k")
            violin_plot[key].set_color(colors)


def _annotate_celltype(ax, n, height, celltype, celltype_color):

    ax.text(
        x=(n + 1),
        y=height,
        s=celltype,
        c=celltype_color,
        ha="center",
        fontsize=8,
        fontfamily="arial",
    )


def _annotate_cell_count(ax, n, depth, n_pts, celltype_color):
    ax.text(
        x=(n + 1),
        y=depth,
        s="n = {}".format(n_pts),
        c=celltype_color,
        ha="center",
        fontsize=6,
        fontfamily="arial",
    )


# Main violin_plot function: --------------------------------------------------
def growth_rate_grouped_violin_plot(adata, ncols=3):
    ViolinPlotDataDict = prepare_data_for_violin_plot(adata)
    fig = ViolinPlot(ncols=ncols, kwargs={"wspace": 0.1})

    for n, (key, dataset) in enumerate(ViolinPlotDataDict.items()):

        title = "{}/{} Growth Rate (Observed) per dominate cell fate".format(
            key.split("_")[0], key.split("_")[1]
        )

        data = list(dataset.values())
        groups = list(dataset.keys())
        panel = fig.axes[n].violinplot(dataset=data)
        colors = [vinplots.colors.LARRY_in_vitro[key] for key in groups]
        _set_violinplot_line_colors(panel, colors)
        ax = fig.axes[n]
        ax.set_title(title, fontsize=10, fontfamily="arial", y=1.05)
        ax.set_ylim(-2, 2)
        for n, body in enumerate(panel["bodies"]):
            celltype = groups[n]
            celltype_color = vinplots.colors.LARRY_in_vitro[celltype]
            n_pts = len(dataset[celltype])
            x_noise = np.random.normal(0, 0.05, n_pts)
            ax.scatter(
                x_noise + (n + 1),
                dataset[celltype],
                c=celltype_color,
                s=25,
                alpha=0.25,
                zorder=2,
            )
            height = max(dataset[celltype]) + 0.15
            depth = min(dataset[celltype]) - 0.15

            _annotate_celltype(ax, n, height, celltype, celltype_color)
            _annotate_cell_count(ax, n, depth, n_pts, celltype_color)

            body.set_facecolor(celltype_color)
            body.set_edgecolor(celltype_color)
            body.set_alpha(0.65)


from torch_adata import AnnDataset
from anndata import AnnData
from typing import List, Union

NoneType = type(None)


def to_AnnDataset(
    adata: AnnData,
    use_key: str = "X_pca",
    groupby: str = "Time point",
    obs_keys: List[str] = None,
    attr_names: {"obs": List[str], "aux": List[str]} = {"obs": [], "aux": []},
    one_hot: List[bool] = False,
    aux_keys: Union[List[str], NoneType] = None,
    silent: bool = False,
    sampling_weight_key: Union[str, NoneType] = None,
):

    return AnnDataset(**locals())




import numpy as np


class Noise:
    """Used to break ties with randomness"""

    def __init__(self, noise_magnitude=0.001):
        self._noise_magnitude = noise_magnitude

    @property
    def _dim1(self):
        return self._df.shape[0]

    @property
    def _dim2(self):
        return self._df.shape[1]

    def forward(self):
        return np.random.rand(self._dim1, self._dim2) * self._noise_magnitude

    def __call__(self, df):
        self._df = df
        self._noise = self.forward()
        return self._df + self._noise



from ._multi_fated_lineage_prediction_results import MultifatedLineagePredictionResults
from ._multiclass_precision_recall import MultiClassPrecisionRecall

from ._negative_cross_entropy import NegativeCrossEntropy
from ._multi_idx_scoring import multi_idx_accuracy, multi_idx_negative_cross_entropy
from ._neutrophil_monocyte_correlation import neutrophil_monocyte_correlation



import scipy
import numpy as np

from ._running_quantile import RunningQuantile


def vscores(E, min_mean=0, nBins=50, fit_percentile=0.1, error_wt=1):
    '''
    Calculate v-score (above-Poisson noise statistic) for genes in the input sparse counts matrix
    Return v-scores and other stats
    '''
    
    runningquantile = RunningQuantile(nBins)

    ncell = E.shape[0]

    mu_gene = E.mean(axis=0).A.squeeze()
    gene_ix = np.nonzero(mu_gene > min_mean)[0]
    mu_gene = mu_gene[gene_ix]

    tmp = E[:,gene_ix]
    tmp.data **= 2
    var_gene = tmp.mean(axis=0).A.squeeze() - mu_gene ** 2
    del tmp
    FF_gene = var_gene / mu_gene

    data_x = np.log(mu_gene)
    data_y = np.log(FF_gene / mu_gene)

    x, y = runningquantile(data_x, data_y, fit_percentile)
    x = x[~np.isnan(y)]
    y = y[~np.isnan(y)]

    gLog = lambda input: np.log(input[1] * np.exp(-input[0]) + input[2])
    h,b = np.histogram(np.log(FF_gene[mu_gene>0]), bins=200)
    b = b[:-1] + np.diff(b)/2
    max_ix = np.argmax(h)
    c = np.max((np.exp(b[max_ix]), 1))
    errFun = lambda b2: np.sum(abs(gLog([x,c,b2])-y) ** error_wt)
    b0 = 0.1
    b = scipy.optimize.fmin(func = errFun, x0=[b0], disp=False)
    a = c / (1+b) - 1


    v_scores = FF_gene / ((1+a)*(1+b) + b * mu_gene);
    CV_eff = np.sqrt((1+a)*(1+b) - 1);
    CV_input = np.sqrt(b);

    return v_scores, CV_eff, CV_input, gene_ix, mu_gene, FF_gene, a, b



import ABCParse


class IndexSubsets(ABCParse.ABCParse):
    """Container for keep track of subset indices."""

    def __init__(self, adata, time_key="Time point", lineage_key="clone_idx"):

        self.__parse__(locals(), public=[None])
        self._df = self._adata.obs.copy()
        self._time = sorted(self._df[self._time_key].unique())
        self._configure_time_subset()
        self._configure_lineage_traced_time_subset()

    @property
    def lineage_traced(self):
        self._lineage_traced = self._df.loc[self._df[self._lineage_key].notna()].index
        return self._lineage_traced

    def _configure_time_subset(self):

        for t in self._time:
            t_idx = self._df.loc[self._df[self._time_key] == t].index
            setattr(self, f"d{int(t)}", t_idx)

    def _configure_lineage_traced_time_subset(self):

        for t in self._time:
            d = getattr(self, f"d{int(t)}")
            d_LT = d.intersection(self.lineage_traced)
            setattr(self, f"d{int(t)}_LT", d_LT)




# -- import packages: ---------------------------------------------------------
import autodevice
import torch
import ABCParse
import pandas as pd
import anndata


# -- import local dependencies: -----------------------------------------------
from ... import utils
from ._fate_bias_matrix_generation import FateBias
from ._observed_fate_bias import F_obs
from .metrics import MultiClassPrecisionRecall

# -- Operational class: -------------------------------------------------------
class FateBiasPredictionTask(ABCParse.ABCParse):
    _F_HAT_CONFIGURED = False
    _F_OBS_CONFIGURED = False
    def __init__(
        self,
        F_hat: pd.DataFrame = None,
        F_obs: pd.DataFrame = None,
        adata: anndata.AnnData = None,
        DiffEq = None,
        kNN_Graph = None,
    ):
        """
        F_hat: pd.DataFrame = None,
        F_obs: pd.DataFrame = None,
        adata: anndata.AnnData = None,
        DiffEq = None,
        kNN_Graph = None,
        """
        self.__parse__(locals(), public = [None])
        
        
    def _configure_F_hat(self, ):
        
        if self._F_hat is None:
            self._F_hat = FateBias(
                self._DiffEq,
                self._kNN_Graph,
            )(self._adata)
        self._F_HAT_CONFIGURED = True
        
    @property
    def F_hat(self):
        if not self._F_HAT_CONFIGURED:
            self._configure_F_hat()
        return self._F_hat
    
    def _configure_F_obs(self, ):
        if self._F_obs is None:
            self._F_obs = F_obs(self._adata).df
        self._F_OBS_CONFIGURED = True
        
    @property
    def F_obs(self):
        if not self._F_OBS_CONFIGURED:
            self._configure_F_obs()
        return self._F_obs
    
    
    @property
    def METRIC_precision_recall(self):
        multiclass_pr = MultiClassPrecisionRecall()
        return = multiclass_pr(self.F_obs, self.F_hat)
    
    
        



import pathlib
import scipy

import ABCParse
from .. import utils

from typing import Union


class ExpressionMatrix(ABCParse.ABCParse):
    def __init__(self, mtx_path: Union[str, pathlib.Path]):

        self.__parse__(locals())

    def read_mtx(self):
        self.mtx = scipy.io.mmread(self.mtx_path)

    @property
    def npz_path(self):
        return pathlib.Path(str(self.mtx_path).replace("mtx.gz", "npz"))

    @property
    def npz_exists(self):
        return self.npz_path.exists()

    def save_mtx_as_npz(self):
        if not self.npz_exists:
            scipy.sparse.save_npz(self.npz_path, self.mtx)


def load_expr_matrix(path: Union[str, pathlib.Path]):
    
    expr_mat = ExpressionMatrix(mtx_path = path)

    if not expr_mat.npz_exists:
        print("Loading expression matrix as `.mtx`. This is slow. Will save as `.npz` for much faster future loading.")
        expr_mat.read_mtx()
        expr_mat.save_mtx_as_npz()
        
    mtx = scipy.sparse.load_npz(expr_mat.npz_path)

    return 




import autodevice
import torch
import sklearn.decomposition
import pandas as pd
import anndata
import lightning

from typing import Optional


def compute_fate_bias(
    adata: anndata.AnnData,
    DiffEq: lightning.LightningModule,
    t0_idx: Optional[pd.Index] = None,
    obs_key: str = "Cell type annotation",
    use_key: str = "X_pca",
    time_key: str = "Time point",
    N: int = 2000,
    kNN: Optional["kNN"] = None,
    PCA: sklearn.decomposition.PCA = None,
    device: torch.device = autodevice.AutoDevice(),
    *args,
    **kwargs,
):
    """Evaluate fate bias

    Args:
        adata (anndata.AnnData): adata obj

        DiffEq (lightning.LightningModule):  Lightning neural differential equation

        t0_idx (Optional[pd.Index]): index of cells from which simulations should be run.
        If not provided, all t0 cells will be used. **Default** = None

        obs_key (str): description. **Default**

        use_key (str): description. **Default**

        time_key (str): description. **Default**

        N (int): number of trajectories to sample for each simulation. **Default** = 2000

        kNN ('kNN'): nearest nieghbors graph.

        PCA (sklearn.decomposition.PCA): **Default** = None

        device (torch.device): **Default** = autodevice.AutoDevice()

    Returns:
        F_hat (pd.DataFrame): predicted fate bias

    Notes:
        1. Default argument values are catered towards the LARRY dataset.
    """

    fate_bias = larry.tasks.fate_prediction.FateBias(device=device)
    function_kwargs = ABCParse.function_kwargs(func=fate_bias.__call__, kwargs=locals())
    return fate_bias(**function_kwargs)





from .... import utils

import pandas as pd
import numpy as np
import ABCParse
import os

import sklearn

class MultiClassPrecisionRecall(ABCParse.ABCParse):
    def __init__(self, threshold=0.3, minor_fate_key="minor"):
        
        self.__parse__(locals(), public=[None])

        self._minor_fate_key = minor_fate_key

        self.precision = {}
        self.recall = {}
        self.mean_precision = {}
        self.AUPR = {}        
                
    def _filter_undiff(self):
        for key in ["undiff", "Undifferentiated"]:
            if key in self._F_obs.columns.tolist():
                self._F_obs = self._F_obs.drop(key, axis = 1)
            if key in self._F_hat.columns.tolist():
                self._F_hat = self._F_hat.drop(key, axis = 1)
                    

    @property
    def _f_hat_fates(self):
        return self._F_hat.columns.tolist()

    @property
    def _f_obs_fates(self):
        return self._F_obs.columns.tolist()

    @property
    def _fates_missing_from_f_hat(self):
        return [i for i in self._f_obs_fates if not i in self._f_hat_fates]

    @property
    def _fates_missing_from_f_obs(self):
        return [i for i in self._f_hat_fates if not i in self._f_obs_fates]

    @property
    def fates(self):
        return list(set(self._f_hat_fates) & set(self._f_obs_fates))

    @property
    def F_obs(self):
        return np.where(self._F_obs > self._threshold, 1, 0)

    def forward(self, i, fate):

        y_true, y_pred = self.F_obs[:, i], self._F_hat[fate]

        (
            self.precision[fate],
            self.recall[fate],
            _,
        ) = sklearn.metrics.precision_recall_curve(y_true, y_pred)
        self.mean_precision[fate] = sklearn.metrics.average_precision_score(
            y_true, y_pred
        )

    def __call__(self, F_obs, F_hat, save_path = None):

        self._F_obs = F_obs
        self._F_hat = F_hat
        
        self._filter_undiff()

        for i, fate in enumerate(self.fates):
            self.forward(i, fate)

        for col in self._fates_missing_from_f_hat:
            self._F_hat[col] = 0

        (
            self.precision[self._minor_fate_key],
            self.recall[self._minor_fate_key],
            _,
        ) = sklearn.metrics.precision_recall_curve(
            self.F_obs.ravel(), self._F_hat.values.ravel()
        )

        self.mean_precision[
            self._minor_fate_key
        ] = sklearn.metrics.average_precision_score(
            self.F_obs, self._F_hat, average="micro"
        )

        for key in self.mean_precision.keys():
            self.AUPR[key] = sklearn.metrics.auc(self.recall[key], self.precision[key])
            
            
        if not save_path is None:
            
            np.save(os.path.join(save_path, "multiclass_pr.recall"), self.recall)
            np.save(os.path.join(save_path, "multiclass_pr.precision"), self.precision)

            pd.DataFrame(
                {
                    "mean_precision": self.mean_precision,
                    "AUPR": self.AUPR,
                }
            ).to_csv(os.path.join(save_path, "mean_precision_aupr.csv"))
            
        return {
            "recall": self.recall,
            "precision": self.precision,
            "mean_precision": self.mean_precision,
            "AUPR": self.AUPR,
        }
    






# -- in vitro data handler: ----------------------------------------------------
class inVitroDataset(DataHandler):
    _url_paths = inVitroURLPaths()
    _dataset = "in_vitro"
    def __init__(self, silent = False):
        super(inVitroData, self).__init__(silent=silent)
        
        
    def fate_prediction(self, split_key="Well", write_h5ad=False):
        
        if self.fate_prediction_h5ad_path.exists():
            print(f"{note} Reading adata prepared for fate prediction task from .h5ad")
            return self.read_h5ad(self.fate_prediction_h5ad_path)
        
        self.adata = self.__call__()
        
        return split_for_fate_prediction_task(
                self.adata,
                split_key=split_key,
                write_h5ad=self.fate_prediction_h5ad_path,
            )

    def timepoint_recovery(self, split_key="Time point"):
        
        if self.timepoint_recovery_h5ad_path.exists():
            print(f"{note} Reading adata prepared for timepoint recovery task from .h5ad")
            return self.read_h5ad(self.timepoint_recovery_h5ad_path)
        
        self.adata = self.__call__()
        
        return split_for_timepoint_recovery_task(
                self.adata,
                split_key=split_key,
                write_h5ad=self.timepoint_recovery_h5ad_path,
            )
    
    def transfer_learning(self, split_key="Time point"):
        
        if self.transfer_learning_h5ad_path.exists():
            print(f"{note} Reading adata prepared for transfer learning task from .h5ad")
            return self.read_h5ad(self.transfer_learning_h5ad_path)
        
        self.adata = self.__call__()
        
        return split_for_transfer_learning_task(
                self.adata,
                split_key=split_key,
                write_h5ad=self.transfer_learning_h5ad_path,
            )



from licorice_font import font_format

    
from ._time_occupance import time_occupance


def fated_idx(
    adata,
    fate_time=[4, 6],
    exclude_fate: tuple = ("Cell type annotation", ["undiff"]),
    lineage_key: str = "clone_idx",
    time_key: str = "Time point",
) -> list:
    """
    generate a list of lineages that are seen at d2 as well as d4 and d6

    Notes:
    ------
    (1) This function returns the indices of lineages NOT cells.
    """
    
    if "time_occupance" in adata.uns_keys():
        t_occ = adata.uns["time_occupance"]
    else:
        t_occ = time_occupance(
            adata,
            fate_time=fate_time,
            exclude_fate=exclude_fate,
            lineage_key=lineage_key,
            time_key=time_key,
            return_df=True,
        )
    
    return t_occ[t_occ["2"] & t_occ[["4", "6"]].any(axis=1)].index.tolist()


def annotate_fated(
    adata,
    lineage_key="clone_idx",
    time_key="Time point",
    t0=2,
    fate_time=[4, 6],
    key_added="fate_observed",
    t0_key_added="t0_fated",
    exclude_fate: tuple = ("Cell type annotation", ["undiff"]),
) -> None:

    """We use this function to denote lineages with cells at d2
    and one or more cells in [d4, d6]
    
    Updates adata.obs with two columns -> adata.obs[['fate_observed', 't0_fated']]
    """

    df = adata.obs.copy()

    f_idx = fated_idx(
        adata,
        fate_time=fate_time,
        exclude_fate=exclude_fate,
        lineage_key=lineage_key,
        time_key=time_key,
    )
    
    info = font_format("INFO", ['PURPLE'])
    df[key_added] = df[lineage_key].isin(f_idx)
    msg = f"- [ {info} ] | Fated cells annotated at: adata.obs['{key_added}']"
    print(msg)
    df[t0_key_added] = (df[time_key] == t0) & df[key_added]
    msg = f"- [ {info} ] | Fated cells (t=t0) annotated at: adata.obs['{t0_key_added}']"
    print(msg)

    adata.obs = df





# import packages: ------------------------------------------------------------
import pandas as pd
from anndata import AnnData


# -----------------------------------------------------------------------------
def count_clonal_lineages(
    adata: AnnData, lineage_key: str = "clone_idx", groupby_key: str = "Time point", return_df: bool = False
) -> pd.DataFrame:

    """
    Count clonal lineages over groups (e.g., time points).

    Parameters:
    -----------
    adata
        type: anndata.AnnData

    lineage_key
        type: str
        default: "clone_idx"

    groupby_key
        type: str
        default: "Time point"

    Returns:
    --------
    count_df

    Notes:
    ------
    """

    meta_df = adata.obs.copy()
    meta_clonal = meta_df.dropna()

    lineage_grouped = meta_clonal.groupby(lineage_key)
    count_df = (
        lineage_grouped[groupby_key].value_counts().unstack().fillna(0).astype(int)
    )
    
    adata.uns['lineage_count_df'] = count_df
    
    if return_df:
        return count_df



# -- import packages: ---------------------------------------------------------------------
import pandas as pd
from licorice_font import font_format
import ABCParse

# -- import local dependencies: -----------------------------------------------------------
from .. import utils
from ._index_subsets import IndexSubsets
from ._annotate_fated import annotate_fated


# -- Main operator class: -----------------------------------------------------------------
class FateValues(ABCParse.ABCParse):
    def __init__(
        self,
        adata,
        origin_time=[2],
        fate_time=[4, 6],
        annotation_key="Cell type annotation",
        time_key="Time point",
        lineage_key="clone_idx",
    ):

        self.__parse__(locals(), public=[None])
        self._df = adata.obs.copy()
        self.idx = IndexSubsets(
            adata, time_key=time_key, lineage_key=lineage_key
        )
        self.keys = {}
        self.keys["lineage"] = self.idx._lineage_key
        self._grouped = adata[self.idx.lineage_traced].obs.groupby(self.keys["lineage"])

    def _count_values_at_lineage_fate(
        self,
        df,
        labels_excluded=["undiff"],
    ):
        """df is the pandas.DataFrame obs table for a single clonal lineage"""
        
        value_key = self._annotation_key
        
        return (
            df.loc[df[self._time_key].isin(self._fate_time)]
            .loc[~df[value_key].isin(labels_excluded)][value_key]
            .value_counts()
        )

    def __call__(self) -> pd.DataFrame:
        """Takes ~13s for the in vitro dataset"""
        return self._grouped.apply(self._count_values_at_lineage_fate).unstack()


def count_t0_cell_fates(adata, key_added="cell_fate_df", return_df=False):
    cell_fate_df = adata.obs[["clone_idx"]].merge(
        adata.uns["fate_counts"], on="clone_idx", how="left"
    )
    cell_fate_df.index = cell_fate_df.index.astype(str)
    
    adata.obsm[key_added] = cell_fate_df

    info = font_format("INFO", ['PURPLE'])
    print(f"- [ {info} ] | Added cell x fate counts to: adata.obsm['{key_added}']")
    if return_df:
        return cell_fate_df


def d2_cell_fate_matrix(
    adata, time_key: str = "Time point", lineage_key: str = "clone_idx",
):
    cell_fate_df = adata.obsm["cell_fate_df"].copy()
    df = adata.obs.copy()
    d2_idx = df.loc[df[time_key] == 2].index
    d2_fate_idx = cell_fate_df.notna().index.isin(d2_idx)  # .reshape(-1, 1)
    cell_fate_df_ = cell_fate_df[d2_fate_idx].dropna()
    return cell_fate_df_[cell_fate_df_.drop(lineage_key, axis=1).sum(1) > 0]


# -- API-facing function: ------------------------------------------------------------
def count_fate_values(
    adata,
    origin_time=[2],
    fate_time=[4, 6],
    annotation_key="Cell type annotation",
    time_key="Time point",
    lineage_key="clone_idx",
    key_added="fate_counts",
    return_dfs=False,
):
    """
    Count fate values.
    """
    fate_values = FateValues(
        adata,
        origin_time=origin_time,
        fate_time=fate_time,
        annotation_key=annotation_key,
        time_key=time_key,
        lineage_key=lineage_key,
    )
    adata.uns[key_added] = lineage_fate_df = fate_values()
    info = font_format("INFO", ['PURPLE'])
    print(f"- [ {info} ] | Added lineage x fate counts to: adata.uns['{key_added}']")

    annotate_fated(
        adata,
        lineage_key=lineage_key,
        time_key=time_key,
        t0=origin_time[0],
        fate_time=fate_time,
        key_added="fate_observed",
        t0_key_added="t0_fated",
        exclude_fate=(annotation_key, ["undiff"]),
                  )
    cell_fate_df = count_t0_cell_fates(adata, return_df=return_dfs)

    d2_cell_fate_df = d2_cell_fate_matrix(adata, time_key = time_key, lineage_key = lineage_key)

    if return_dfs:
        return lineage_fate_df, cell_fate_df, d2_cell_fate_df




# -- import packages: ------------------------------------
import larry
import numpy as np
import pandas as pd
import seaborn as sns


# -- import local dependencies: --------------------------
from ._time_occupance import time_occupance


# -- supporting functions: -------------------------------
def _sum_normalize_fate(fate_df: pd.DataFrame) -> pd.DataFrame:

    """
    Given a DataFrame of integer counted fate/state observations, normalize to the total within the sample.

    Parameters:
    -----------
    fate_df
        DataFrame with fates along the columns and lineages along the index.
        type: pandas.DataFrame

    Notes:
    ------
    (1) If not already filtered for null-valued fates at the time of observations, these will return NaNs due
        to the resulting div. by zero.
    (2) Can have a multi-index but select for only a single timepoint if that's the desired behavior.
    """

    norm_fates = fate_df.values / fate_df.sum(1).values[:, None]
    return pd.DataFrame(norm_fates, index=fate_df.index, columns=fate_df.columns)


def state_description_at_time(df, time_key="Time point", state_label_key="Annotation"):
    return df.groupby(time_key)[state_label_key].value_counts().unstack()

# -- main class: -----------------------------------------

class CellFates:
    def __init__(
        self,
        adata,
        time_key="Time point",
        lineage_key="clone_idx",
        state_label_key="Annotation",
    ):

        self.__setup__(locals())

    def __parse__(self, kwargs, ignore=["self"]):

        for key, val in kwargs.items():
            if not key in ignore:
                setattr(self, key, val)

    def __setup__(self, kwargs, ignore=["self"]):

        self.__parse__(kwargs, ignore)
        self.df = self.adata.obs.copy()
        self.time_occ = larry.lt.time_occupance(self.adata, return_df=True)
        self.t = sorted(self.adata.obs[self.time_key].unique())

    def subset(self, subset_time=None):

        """
        # subset for lineages with cells at d2 and d6
        # cf.subset([2, 6])
        # subset for lineages with cells at all time points
        """

        if not subset_time:
            subset_time = self.t

        self.subset_time = subset_time
        self.subset_lineages = self.time_occ.loc[self.time_occ[self.subset_time].all(1)]
        self.subset_lin_idx = self.subset_lineages.index
        self.n_lineages = len(self.subset_lin_idx)
        self.adata_subset = self.adata[
            self.df.loc[self.df[self.lineage_key].isin(self.subset_lin_idx)].index
        ].copy()
        self.df_subset = self.adata_subset.obs.copy()
        self.d2_df = self.df.loc[self.df["Time point"] == 2]

    def _lineage_state_descriptions(self):

        self.lineage_states = self.df_subset.groupby(self.lineage_key).apply(
            state_description_at_time, self.time_key, self.state_label_key
        )

    def _configure_fates(self):
        self._fates = sorted(self.df[self.state_label_key].unique())

    @property
    def fates(self):
        if not hasattr(self, "_fates"):
            self._configure_fates()
        return self._fates

    @property
    def lineage_state_descriptions(self) -> pd.DataFrame:
        if not hasattr(self, "lineage_states"):
            self._lineage_state_descriptions()
        return self.lineage_states

    def temporal_fate_descriptions(self):

        for n, _t in enumerate(self.t):
            if _t in self.subset_time:
                tmp_df = self.lineage_state_descriptions.iloc[n :: len(self.t)]
                tmp_df = pd.DataFrame(
                    tmp_df.values, index=self.subset_lin_idx, columns=self.fates
                )
                setattr(self, "_raw_d{}_state".format(int(_t)), tmp_df)
                setattr(self, "d{}_state".format(int(_t)), _sum_normalize_fate(tmp_df))

    def combine_fate_matrices(self, t_combine: list = [4, 6], normalize=True):
        """Combine fate matrices across timepoints (e.g., d4 and d6)"""

        states = [getattr(self, "_raw_d{}_state".format(t)) for t in t_combine]
        combined_states = np.zeros(states[0].shape)
        for state in states:
            combined_states += state.values

        self.combined_state_df = _sum_normalize_fate(
            pd.DataFrame(combined_states, index=self.subset_lin_idx, columns=self.fates)
        )

    def clustermap(self, fate_df):
        """cf.clustermap(cf.combined)"""
        self.cg = sns.clustermap(
            fate_df, cmap="Purples", figsize=(5, 5), yticklabels=False
        )

    def format_state_for_merge(self, state_df):
        return state_df.reset_index().rename({"index": "clone_idx"}, axis=1)

    def state_to_cell_indexed_fate_bias(self, state_df):
        """
        Transform cell state df to a cell-indexed fate bias df

        Can pass: cf.d6_state, cf.d4_state
        """

        idx_df = self.adata.obs.copy().filter(regex="idx").dropna()
        state_df = (
            state_df.reset_index().rename({"index": "clone_idx"}, axis=1).dropna()
        )
        fates = state_df.columns.tolist()[1:]
        df = pd.merge(idx_df, state_df, on="clone_idx")
        df.index = df["cell_idx"].values

        return df[fates]

    def __call__(self, subset_time=[2, 6], fate_t=6):

        self.subset(subset_time=subset_time)
        self.temporal_fate_descriptions()
        state_df = getattr(self, "d{}_state".format(int(fate_t)))
        return self.state_to_cell_indexed_fate_bias(state_df)  # fate_df




__module_name__ = "__init__.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# -----------------------------------------------------------------------------
from ._count_clonal_lineages import count_clonal_lineages
from ._count_d2_daughter_cells import count_d2_daughter_cells
from ._count_clones_by_timepoint import count_clones_by_timepoint
from ._estimate_growth_rates import estimate_growth_rates
from ._estimate_growth_rates import plot_growth_rates

from ._get_lineage_obs import get_lineage_obs
from ._calculate_dominate_fate import calculate_dominate_fate
from ._get_annotated_metadata_d2_lineage_cells import get_annotated_metadata_d2_lineage_cells
# from ._growth_rate_grouped_violin_plot import growth_rate_grouped_violin_plot
from ._LARRY_t0_test_indices import LARRY_t0_test_indices
# from ._temporal_cmap import temporal_cmap



# import scdiffeq_analyses as sdq_an
# import numpy as np

# NoneType = type(None)


# def temporal_cmap(t=None, idx_slice=1, pad_left=1, pad_right=1):

#     t_cmap = sdq_an.pl.TimeColorMap()(
#         idx_slice=idx_slice, pad_left=pad_left, pad_right=pad_right
#     )
#     if isinstance(t, NoneType):
#         return t_cmap
#     else:
#         return t_cmap[np.floor(np.linspace(0, len(t_cmap) - 1, len(t))).astype(int)]



# -- import packages: ----------------------------------------------------------
from pathlib import Path
import pandas as pd
import numpy as np
import anndata
import ABCParse
import scipy
import scipy.io

# -- import local dependencies: ------------------------------------------------
from .. import utils
from ._dataset_utils import Messages

# -- supporting functions and classes: -----------------------------------------
def _read_format_df(path, **kwargs):
    df = pd.read_csv(path, **kwargs)
    df.index = df.index.astype(str)
    return df

class ExpressionMatrix(ABCParse.ABCParse):
    """Container for parsing the expression matrix."""

    def __init__(self, mtx_path: str):

        self.__parse__(locals())
        

    def read_mtx(self)->None:
        self.mtx = scipy.io.mmread(self.mtx_path)

    @property
    def npz_path(self)->Path:
        return Path(str(self.mtx_path).replace("mtx.gz", "npz"))

    @property
    def npz_exists(self)->bool:
        return self.npz_path.exists()

    def save_mtx_as_npz(self)->None:
        if not self.npz_exists:
            scipy.sparse.save_npz(self.npz_path, self.mtx)

    def __call__(self) -> scipy.sparse.csr_matrix:

        """
        Load the expression matrix.

        Parameters:
        -----------
        None

        Returns:
        --------
        X
            type: scipy.sparse.csr_matrix

        Given a path to a .mtx file, initializes the ExpressionMatrix class.
        Looks for the .npz file. If it's not there, the .mtx file is parsed
        and saved as .npz.
        """
        if not self.npz_exists:
            messages.mtx_to_npz()
            self.read_mtx()
            self.save_mtx_as_npz()

        return scipy.sparse.csr_matrix(scipy.sparse.load_npz(self.npz_path))


class AnnDataConfiguration(ABCParse.ABCParse):
    """Construct AnnData from constituent components."""

    def __init__(self, X_path, obs_path, var_path, X_clone_path, silent=False):

        """Pass paths of adata components."""

        self.__parse__(locals(), public=[None])
        self._expr_mtx = ExpressionMatrix(mtx_path=self._X_path)

    @property
    def X(self)->scipy.sparse.csr_matrix:
        """returns expression matrix"""
        if not hasattr(self, "_X"):
            self._X = self._expr_mtx()
        return self._X

    @property
    def var(self)->pd.DataFrame:
        """returns var pd.DataFrame"""
        if not hasattr(self, "_var_df"):
            self._var_df = _read_format_df(
                self._var_path, header=None, names=["gene_ids"]
            )
        return self._var_df

    @property
    def obs(self)->pd.DataFrame:
        """returns obs pd.DataFrame"""
        if not hasattr(self, "_obs_df"):
            self._obs_df = _read_format_df(self._obs_path, sep="\t")
        return self._obs_df

    @property
    def X_clone(self)->scipy.sparse.csr_matrix:
        """returns cell x clonal barcode matrix"""
        if not hasattr(self, "_X_clone"):
            self._X_clone = scipy.sparse.csr_matrix(scipy.io.mmread(self._X_clone_path))
            # make sure it is oriented correctly
            if self._X_clone.shape[0] != self.X.shape[0]:
                self._X_clone = self._X_clone.T
        return self._X_clone

    @property
    def adata(self):
        """returns formatted AnnData object"""
        if not hasattr(self, "_adata"):
            self._adata = anndata.AnnData(
                X=self.X,
                dtype=self.X.dtype,
                obs=self.obs,
                var=self.var,
                obsm={"X_clone": self.X_clone},
            )
        return self._adata




# -- import packages: ----------------------------------------------------------
import os
from pathlib import Path
from licorice_font import font_format


# -- Main class: ---------------------------------------------------------------
class DirectoryManager:
    def __init__(self, path: str, silent: bool = False):

        self._path = path
        self._dirs_created = []
        self._silent = silent

    @property
    def split_path(self):
        return [subdir for subdir in self._path.split("/") if subdir]    

    def notify(self, path):
            msg = font_format(f"mkdir: {path}", ['BLUE'])
            print(f" - [NOTE] | {msg}")

    def __mkdir__(self, path):
        path.mkdir()
        self._dirs_created.append(path)
        if not self._silent:
            self.notify(path)


    def __call__(self):

        target_dir = [""]

        for subdir in self.split_path:
            target_dir.append(subdir)
            subdir_ = Path("/".join(target_dir))
            if not subdir_.exists():
                self.__mkdir__(subdir_)


# -- API-facing function: ------------------------------------------------------
def mkdir(path: str, silent: bool = False)->None:

    if not Path(path).is_dir():
        path = os.path.dirname(path)

    dir_manager = DirectoryManager(path = path, silent = silent)
    dir_manager()



import setuptools
import re
import os
import sys

setuptools.setup(
    name='LARRY-dataset',
    version='0.0.2rc1',
    python_requires='>3.9.0',
    author='Michael E. Vinyard',
    author_email='mvinyard.ai@gmail.com',
    url = 'https://github.com/scDiffEq/LARRY-dataset',
    long_description = open("README.md", encoding="utf-8").read(),
    long_description_content_type = 'text/markdown',
    description='LARRY Dataset: lineage and RNA recovery',
    packages = setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
        "anndata>=0.9.1",
  #      "pytorch-lightning>=2.0.2",
   #     "wget>=3.2",
    ],
    classifiers = [
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3.9",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    license="MIT"
)




# -- import packages: ----------------------------------------------------------
import anndata


# -- API-facing function: ------------------------------------------------------
def subset_clonal(adata: anndata.AnnData, lineage_key: str = "clone_idx"):

    """
    Parameters:
    -----------
    adata
        type: anndata.AnnData

    lineage_key
        type: str
        default: "clone_idx"

    Returns:
    --------
    clonal_adata
    """

    df = adata.obs.copy()
    clonal_adata = adata[df.loc[df[lineage_key].notna()].index].copy()
    clonal_adata.obs = clonal_adata.obs.reset_index()
    clonal_adata.obs.index = clonal_adata.obs.index.astype(str)

    return clonal_adata


import os
from pathlib import Path
import licorice_font


class AnnDataPathManager:
    def __init__(self, dataset, download_dir):

        self._dataset = dataset
        self._download_dir = os.path.join(download_dir, f"KleinLabData/{self._dataset}")

    def _path_constructor(self, obj_specifier):
        fpath = f"adata.Weinreb2020.{self._dataset}.{obj_specifier}.h5ad"
        return Path(os.path.join(self._download_dir, fpath))

    @property
    def raw(self):
        return self._path_constructor("raw")

    @property
    def gene_filtered(self):
        return self._path_constructor("gene_filtered")

    @property
    def timepoint_recovery(self):
        return self._path_constructor("task_01.timepoint_recovery")

    @property
    def fate_prediction(self):
        return self._path_constructor("task_02.fate_prediction")

    @property
    def uniform(self):
        return self._path_constructor("uniform")

    def __repr__(self):

        description = "{}:\n".format(
            licorice_font.font_format("AnnData Objects", ["BOLD"])
        )
        description += "----------------"

        attrs = {
            attr: getattr(self, attr)
            for attr in self.__dir__()
            if not attr.startswith("_")
        }

        for k, v in attrs.items():
            if v.exists():
                exists = "[{:^20}]".format(
                    licorice_font.font_format(str(v.exists()), ["BOLD", "BLUE"])
                )
            else:
                exists = "[{:^15}]".format(
                    licorice_font.font_format(str(v.exists()), ["BOLD"])
                )
            description += f"\n- {exists} | adata.Weinreb2020.{self._dataset}.{k}.h5ad"

        return description